'''
Stanley Bak

Implementation of concrete aggregation methods, used by aggdag
'''

import numpy as np

from hylaa.stateset import StateSet
from hylaa import lputil

def aggregate_box_arnoldi(agg_list, op_list, is_box, is_arnoldi, add_guard, print_func):
    '''
    perform template-based aggregation on the passed-in list of states

    Currently, this can either use box template directions or arnoldi (+box) template directions
    '''

    assert is_box or is_arnoldi

    min_step = min([state.cur_steps_since_start[0] for state in agg_list])
    max_step = max([state.cur_steps_since_start[1] for state in agg_list])
    step_interval = [min_step, max_step]

    print_func("Aggregation time step interval: {}".format(step_interval))

    # create a new state from the aggregation
    postmode = agg_list[0].mode
    postmode_dims = postmode.a_csr.shape[0]
    mid_index = len(agg_list) // 2

    op = op_list[mid_index]

    if is_box or op is None:
        agg_dir_mat = np.identity(postmode_dims)
    elif is_arnoldi:
        # aggregation with a predecessor, use arnoldi directions in predecessor mode in center of
        # middle aggregagted state, then project using the reset, and reorthogonalize

        premode = op.parent_node.stateset.mode
        t = op.transition
        print_func("aggregation point: {}".format(op.premode_center))

        premode_dir_mat = lputil.make_direction_matrix(op.premode_center, premode.a_csr)
        print_func("premode dir mat:\n{}".format(premode_dir_mat))

        if t.reset_csr is None:
            agg_dir_mat = premode_dir_mat
        else:
            projected_dir_mat = premode_dir_mat * t.reset_csr.transpose()

            print_func("projected dir mat:\n{}".format(projected_dir_mat))

            # re-orthgohonalize (and create new vectors if necessary)
            agg_dir_mat = lputil.reorthogonalize_matrix(projected_dir_mat, postmode_dims)

        # also add box directions in target mode (if they don't already exist)
        box_dirs = []
        for dim in range(postmode_dims):
            direction = [0 if d != dim else 1 for d in range(postmode_dims)]
            exists = False

            for row in agg_dir_mat:
                if np.allclose(direction, row):
                    exists = True
                    break

            if not exists:
                box_dirs.append(direction)

        if box_dirs:
            agg_dir_mat = np.concatenate((agg_dir_mat, box_dirs), axis=0)

    if op and add_guard:
        # add all the guard conditions to the agg_dir_mat
        
        t = op.transition

        if t.reset_csr is None: # identity reset
            guard_dir_mat = t.guard_csr
        else:
            # multiply each direction in the guard by the reset
            guard_dir_mat = t.guard_csr * t.reset_csr.transpose()

        if guard_dir_mat.shape[0] > 0:
            agg_dir_mat = np.concatenate((agg_dir_mat, guard_dir_mat.toarray()), axis=0)

    print_func("agg dir mat:\n{}".format(agg_dir_mat))
    lpi_list = [state.lpi for state in agg_list]

    new_lpi = lputil.aggregate(lpi_list, agg_dir_mat, postmode)

    return StateSet(new_lpi, agg_list[0].mode, step_interval, op_list, is_concrete=False)

def aggregate_chull(agg_list, op_list, print_func):
    '''
    perform template-based aggregation on the passed-in list of states

    Currently, this can either use box template directions or arnoldi (+box) template directions
    '''

    min_step = min([state.cur_steps_since_start[0] for state in agg_list])
    max_step = max([state.cur_steps_since_start[1] for state in agg_list])
    step_interval = [min_step, max_step]

    print_func("Convex hull aggregation time step interval: {}".format(step_interval))

    postmode = agg_list[0].mode
    lpi_list = [state.lpi for state in agg_list]

    new_lpi = lputil.aggregate_chull(lpi_list, postmode)

    return StateSet(new_lpi, agg_list[0].mode, step_interval, op_list, is_concrete=False)



'''
Stanley Bak
Functions related to plotting a set of lp constriants

get_verts is probably the main one to use

make_plot_vecs is useful for controlling the accuracy (and decreasing overhead compared w/ not passing it to get_verts)
'''

import math
import numpy as np
from scipy.spatial import ConvexHull

import hylaa.kamenev as kamenev
from hylaa.timerutil import Timers

def pt_to_plot_xy(pt, xdim=0, ydim=1, cur_time=0.0):
    '''convert a point to an x/y pair for plotting
    xdim and ydim can be either an integer (dimension number), an np.array (direction), or None (time will be used)
    '''

    x = y = float(cur_time)

    if isinstance(xdim, int):
        x = pt[xdim]
    elif xdim:
        assert isinstance(xdim, np.ndarray)
        x = np.dot(pt, xdim)

    if ydim and isinstance(ydim, int):
        y = pt[ydim]
    elif ydim:
        assert isinstance(ydim, np.ndarray)
        y = np.dot(pt, ydim)

    return x, y

def get_verts(lpi, xdim=0, ydim=1, plot_vecs=None, cur_time=0.0):
    '''get the vertices defining (an underapproximation) of the outside of the given linear constraints
    These will be usable for plotting, so that rv[0] == rv[-1]. A single point may be returned if the constraints
    are (close to) a single point.

    xdim and ydim can be either an integer (dimension number), an np.array (direction), or None (time will be used)

    cur_time can be an integer or a list of length 2 (min/max time)

    plot_vecs is an ordered list of vectors defining all the 2-d directions to optimize in... if None will 
    construct and use 256 equally spaced vectors 
    '''

    tol = 1e-9
    
    if plot_vecs is None:
        plot_vecs = make_plot_vecs()

    # make cur_time a 2-tuple: (min_time, max time)
    if isinstance(cur_time, (float, int)):
        cur_time = [float(cur_time), float(cur_time)]

    #try:
    if xdim is None and ydim is None:
        if abs(cur_time[0] - cur_time[1]) < tol:
            pts = [[cur_time, cur_time]]
        else:
            pts = []
            pts.append([cur_time.min, cur_time.min])
            pts.append([cur_time.min, cur_time.max])
            pts.append([cur_time.max, cur_time.max])
            pts.append([cur_time.max, cur_time.min])

    elif xdim is None:
        # plot over time
        if isinstance(ydim, int):
            ydim = np.array([1.0 if dim == ydim else 0.0 for dim in range(lpi.dims)], dtype=float)

        lpi.set_minimize_direction(ydim)
        res = lpi.minimize(columns=[lpi.cur_vars_offset + n for n in range(lpi.dims)])
        ymin = np.dot(ydim, res)

        lpi.set_minimize_direction(-1 * ydim)
        res = lpi.minimize(columns=[lpi.cur_vars_offset + n for n in range(lpi.dims)])
        ymax = np.dot(ydim, res)

        verts = [[cur_time[0], ymin]]

        if abs(ymin - ymax) > tol:
            verts.append([cur_time[0], ymax])

            if abs(cur_time[0] - cur_time[1]) > tol:
                verts.append([cur_time[1], ymax])

        if abs(cur_time[0] - cur_time[1]) > tol:
            verts.append([cur_time[1], ymin])

    elif ydim is None:
        # plot over time
        if isinstance(xdim, int):
            xdim = np.array([1.0 if dim == xdim else 0.0 for dim in range(lpi.dims)], dtype=float)

        lpi.set_minimize_direction(xdim)
        res = lpi.minimize(columns=[lpi.cur_vars_offset + n for n in range(lpi.dims)])
        xmin = np.dot(xdim, res)

        lpi.set_minimize_direction(-1 * xdim)
        res = lpi.minimize(columns=[lpi.cur_vars_offset + n for n in range(lpi.dims)])
        xmax = np.dot(xdim, res)

        verts = [[xmin, cur_time[0]]]

        if abs(xmin - xmax) > tol:
            verts.append([xmax, cur_time[0]])

            if abs(cur_time[0] - cur_time[1]) > tol:
                verts.append([xmax, cur_time[1]])

        if abs(cur_time[0] - cur_time[1]) > tol:
            verts.append([xmin, cur_time[1]])
    else:
        # 2-d plot
        bboxw = bbox_widths(lpi, xdim, ydim)

        # use bbox to determine acceptable accuracy... might be better to somehow do this
        epsilon = min(bboxw) / 1000.0
        dim_list = [xdim, ydim]

        def supp_point_nd(vec):
            'return a supporting point for the given direction (maximize)'

            assert len(vec) == len(dim_list)

            d = np.zeros((lpi.dims,), dtype=float)
            # negative here because we want to MAXIMIZE not minimize

            for i, dim_index in enumerate(dim_list):
                d[dim_index] = -vec[i]

            lpi.set_minimize_direction(d)

            res = lpi.minimize(columns=[lpi.cur_vars_offset + n for n in range(lpi.dims)])

            rv = []

            for dim in dim_list:
                rv.append(res[dim])

            rv = np.array(rv, dtype=float)

            return rv

        Timers.tic('kamenev.get_verts')
        verts = kamenev.get_verts(len(dim_list), supp_point_nd, epsilon=epsilon)
        Timers.toc('kamenev.get_verts')

        if len(verts) > 2:
            # make 2-d convex hull to fix the order
            hull = ConvexHull(verts)

            verts = [[verts[i, 0], verts[i, 1]] for i in hull.vertices]

        # old method
        #pts = find_boundary_pts(lpi, xdim, ydim, plot_vecs, bboxw)
        #verts = [[pt[0], pt[1]] for pt in pts]

    # wrap polygon back to first point
    verts.append(verts[0])

    return verts

def bbox_widths(lpi, xdim, ydim):
    'find and return the bounding box widths of the lp for the passed-in dimensions'

    rv = []
    dims = lpi.dims

    for dim in [xdim, ydim]:
        col = lpi.cur_vars_offset + dim
        min_dir = [1 if i == dim else 0 for i in range(dims)]
        max_dir = [-1 if i == dim else 0 for i in range(dims)]
        
        min_val = lpi.minimize(direction_vec=min_dir, columns=[col])[0]
        max_val = lpi.minimize(direction_vec=max_dir, columns=[col])[0]

        dx = max_val - min_val

        if dx < 1e-5:
            dx = 1e-5

        rv.append(dx)

    return rv

def make_plot_vecs(num_angles=256, offset=0.0):
    'make plot_vecs with equally spaced directions, returning the result'

    plot_vecs = []

    for theta in np.linspace(0.0, 2.0*math.pi, num_angles, endpoint=False):
        x = math.cos(theta + offset)
        y = math.sin(theta + offset)

        vec = np.array([x, y], dtype=float)

        plot_vecs.append(vec)

    return plot_vecs

def find_boundary_pts(lpi, xdim, ydim, plot_vecs, bboxw):
    '''
    find points along an LPs boundary by solving several LPs and
    returns a list of points on the boundary which maximize each
    of the passed-in directions
    '''

    rv = []

    assert len(plot_vecs) >= 2

    # optimized approach: do binary search to find changes
    point = _minimize(lpi, xdim, ydim, plot_vecs[0], bboxw)
    rv.append(point.copy())

    # add it in thirds, to ensure we don't miss anything
    third = len(plot_vecs) // 3

    # 0 to 1/3
    point = _minimize(lpi, xdim, ydim, plot_vecs[third], bboxw)

    if not np.array_equal(point, rv[-1]):
        rv += _binary_search_boundaries(lpi, 0, third, rv[-1], point, xdim, ydim, plot_vecs, bboxw)
        rv.append(point.copy())

    # 1/3 to 2/3
    point = _minimize(lpi, xdim, ydim, plot_vecs[2*third], bboxw)

    if not np.array_equal(point, rv[-1]):
        rv += _binary_search_boundaries(lpi, third, 2*third, rv[-1], point, xdim, ydim, plot_vecs, bboxw)
        rv.append(point.copy())

    # 2/3 to end
    point = _minimize(lpi, xdim, ydim, plot_vecs[-1], bboxw)

    if not np.array_equal(point, rv[-1]):
        rv += _binary_search_boundaries(lpi, 2*third, len(plot_vecs) - 1, rv[-1], point, xdim, ydim, plot_vecs, bboxw)
        rv.append(point.copy())

    # pop last point if it's the same as the first point
    if len(rv) > 1 and np.array_equal(rv[0], rv[-1]):
        rv.pop()

    return rv

def _minimize(lpi, xdim, ydim, direction, bounding_box_widths):
    'minimize to lp... returning the 2-d point of the minimum'

    dims = 0

    assert not (xdim is None and ydim is None)

    if xdim is not None:
        dims = xdim + 1 if isinstance(xdim, int) else len(xdim)

    if ydim is not None:
        dims = max(dims, ydim + 1) if isinstance(ydim, int) else max(dims, len(ydim))
        
    if isinstance(xdim, int):
        xdim = np.array([1.0 if dim == xdim else 0.0 for dim in range(dims)], dtype=float)

    if isinstance(ydim, int):
        ydim = np.array([1.0 if dim == ydim else 0.0 for dim in range(dims)], dtype=float)

    # xdim and ydim are both arrays or None
    optimize_direction = np.zeros(dims, dtype=float)

    if xdim is not None:
        optimize_direction += direction[0] * xdim / bounding_box_widths[0]

    if ydim is not None:
        optimize_direction += direction[1] * ydim / bounding_box_widths[1]

    lpi.set_minimize_direction(optimize_direction)

    res = lpi.minimize(columns=[lpi.cur_vars_offset + n for n in range(dims)])

    xcoord = 0 if xdim is None else np.dot(res, xdim)
    ycoord = 0 if ydim is None else np.dot(res, ydim)
    
    return np.array([xcoord, ycoord], dtype=float)

def _binary_search_boundaries(lpi, start, end, start_point, end_point, xdim, ydim, plot_vecs, bboxw):
    '''
    return all the optimized points in the star for the passed-in directions, between
    the start and end indices, exclusive

    points which match start_point or end_point are not returned
    '''

    rv = []

    if start + 1 < end:
        mid = (start + end) // 2

        mid_point = _minimize(lpi, xdim, ydim, plot_vecs[mid], bboxw)

        not_start = not np.allclose(start_point, mid_point, atol=1e-6)
        not_end = not np.allclose(end_point, mid_point, atol=1e-6)

        if not_start:
            rv += _binary_search_boundaries(lpi, start, mid, start_point, mid_point, xdim, ydim, plot_vecs, bboxw)

        if not_start and not_end:
            rv.append(mid_point)

        if not_end:
            rv += _binary_search_boundaries(lpi, mid, end, mid_point, end_point, xdim, ydim, plot_vecs, bboxw)

    return rv



'''
Stanley Bak
Hylaa-Continuous Guard Optimization Logic
July 2017
'''

import numpy as np

from scipy.sparse import csc_matrix

from hylaa.util import Freezable
from hylaa.hybrid_automaton import LinearAutomatonMode
from hylaa.settings import PlotSettings
from hylaa.glpk_interface import LpInstance
from hylaa.timerutil import Timers

class GuardOptData(Freezable):
    'Guard optimization data'

    def __init__(self, star, mode, transition_index):
        assert isinstance(mode, LinearAutomatonMode)

        self.settings = star.settings
        self.mode = mode
        self.inputs = star.inputs

        self.star = star
        self.transition = mode.transitions[transition_index]
        self.num_output_vars = self.transition.guard_matrix_csr.shape[1]

        self.key_dir_offset = 0

        if self.settings.plot.plot_mode != PlotSettings.PLOT_NONE:
            if self.settings.plot.xdim_dir is not None:
                self.key_dir_offset += 1

            if self.settings.plot.ydim_dir is not None:
                self.key_dir_offset += 1

        self.lpi = LpInstance(self.num_output_vars, star.num_init_vars, self.inputs)

        self.lpi.set_init_constraints(star.init_mat, star.init_rhs)
        self.lpi.set_output_constraints(self.transition.guard_matrix_csr, self.transition.guard_rhs)

        if star.inputs > 0:
            self.lpi.set_input_constraints_csc(csc_matrix(star.mode.u_constraints_csr), star.mode.u_constraints_rhs)

        self.optimized_lp_solution = None

        self.freeze_attrs()

    def update_full_lp(self):
        '''update the LP solution and, if it's feasible, get its solution, for GUARD_FULL_LP'''

        cur_basis_mat = self.star.time_elapse.cur_basis_mat

        # start and end rows of key-dir matrices
        start = self.key_dir_offset
        end = self.key_dir_offset + self.num_output_vars

        self.lpi.update_basis_matrix(cur_basis_mat[start:end])

        # add input effects for the current step (if it exists)
        if self.star.time_elapse.cur_input_effects_matrix is not None:
            input_effects_mat = self.star.time_elapse.cur_input_effects_matrix
            self.lpi.add_input_effects_matrix(input_effects_mat[start:end])

        result_len = self.num_output_vars + self.star.num_init_vars
        result_len += self.num_output_vars # total input effect
        result_len += self.inputs * (self.star.time_elapse.next_step - 1) # inputs at each step

        result = np.zeros((result_len), dtype=float)
        direction = np.zeros((self.num_output_vars,), dtype=float)

        is_feasible = self.lpi.minimize(direction, result, error_if_infeasible=False)

        return result if is_feasible else None

    def get_guard_lpi(self):
        '''get the current full lp instance for this guard'''

        return self.lpi

    def get_optimized_lp_solution(self):
        '''gets the lp solution without calling an lp solver
        this is only possible if the initial set has only range conditions for each initial dimension,
        and the output-space has only a single condition

        This returns either an lp solution (np.ndarray) or None if infeasible
        '''

        Timers.tic('get_optimized_lp_solution')

        init_ranges = self.star.init_range_tuples
        basis_mat = self.star.time_elapse.cur_basis_mat
        input_effects_mat = self.star.time_elapse.cur_input_effects_matrix

        assert len(init_ranges) == basis_mat.shape[1]
        assert len(self.transition.guard_rhs) == 1
        assert self.num_output_vars == 1
        assert self.transition.guard_matrix_csr.shape == (1, 1)

        guard_multiplier = self.transition.guard_matrix_csr[0, 0]

        if self.optimized_lp_solution is None:
            # +1 for output +1 for total input effects
            self.optimized_lp_solution = [0.0] * (len(init_ranges) + 1 + 1)

        result = self.optimized_lp_solution
        total_output_index = len(init_ranges)
        total_input_index = len(init_ranges) + 1

        guard_threshold = self.transition.guard_rhs[0]
        noinput_effect = 0

        Timers.tic("noinput_effects")
        for init_index in xrange(len(init_ranges)):
            basis_val = basis_mat[0, init_index]
            min_init = init_ranges[init_index][0]
            max_init = init_ranges[init_index][1]

            mult = basis_val * guard_multiplier
            val1 = min_init * mult
            val2 = max_init * mult

            # take the minimum of val1 and val2, since guard is CONDITION <= RHS
            if val1 < val2:
                noinput_effect += val1
                result[init_index] = min_init
            else:
                noinput_effect += val2
                result[init_index] = max_init
        Timers.toc("noinput_effects")

        # add input effects if they exist
        if input_effects_mat is not None:
            Timers.tic("input_effects")
            input_ranges = self.star.mode.u_range_tuples

            for input_index in xrange(len(input_ranges)):
                basis_val = input_effects_mat[0][input_index]
                min_input = input_ranges[input_index][0]
                max_input = input_ranges[input_index][1]

                val1 = min_input * basis_val * guard_multiplier
                val2 = max_input * basis_val * guard_multiplier

                # take the minimum of val1 and val2, since guard is CONDITION <= RHS
                if val1 < val2:
                    result[total_input_index] += val1
                    result.append(min_input)
                else:
                    result[total_input_index] += val2
                    result.append(max_input)

            Timers.toc("input_effects")

        result[total_output_index] = noinput_effect + result[total_input_index]

        rv = np.array(result, dtype=float) if result[total_output_index] <= guard_threshold else None

        Timers.toc('get_optimized_lp_solution')

        return rv

    def get_updated_lp_solution(self):
        '''update the LP solution and, if it's feasible, get its solution'''

        if self.star.settings.interval_guard_optimization and self.star.init_range_tuples is not None and \
            (self.star.inputs == 0 or self.star.mode.u_range_tuples is not None) and self.num_output_vars == 1:
            # LP can be decomposed column-by-column (optimization)
            rv = self.get_optimized_lp_solution()
        else:
            rv = self.update_full_lp()

        return rv

def compute_noinput_effects_split(param):
    'compute noinput effects for part of the basis matrix (used for multithreaded computation)'

    start, end, init_ranges, basis_mat, guard_multiplier = param

    assert len(init_ranges) == basis_mat.shape[1]

    result = np.zeros((len(init_ranges),), dtype=float)
    noinput_effect = 0

    

    return result, noinput_effect



'''This file defines which files to import when 'from hylaa import *' is used'''

__all__ = []

from . import _version
__version__ = _version.get_versions()['version']
__author__ = "Stanley Bak"
__copyright__ = "Copyright 2023"
__credits__ = [
   "Stanley Bak",
   "Hoang-Dung Tran",
   "Max von Hippel" 
]
__license__ = "GPLv3"
__maintainer__ = "Stanley Bak"
__email__ = "bak2007-DONTSENDMEEMAIL@gmail.com"
__status__ = "Prototype"


'''
Time-elapse object for matrix exponential and expm-mul methods

Stanley Bak
April 2018
'''

from scipy.sparse import csc_matrix
from scipy.sparse.linalg import expm, expm_multiply

import numpy as np

from hylaa.util import Freezable
from hylaa.timerutil import Timers

class TimeElapseExpmMult(Freezable):
    'container object for expm + matrix-vec mult method'

    def __init__(self, time_elapser):
        self.time_elapser = time_elapser
        self.a_csc = csc_matrix(time_elapser.mode.a_csr)
        self.b_csc = None if time_elapser.mode.b_csr is None else csc_matrix(time_elapser.mode.b_csr)
        self.dims = time_elapser.dims

        self.cur_step = 0
        self.cur_basis_matrix = None
        self.cur_input_effects_matrix = None

        self.one_step_matrix_exp = None # one step matrix exponential
        self.one_step_input_effects_matrix = None # one step input effects matrix, if inputs exist

        # lgg approximation model vars
        self.use_lgg = False

        self.freeze_attrs()

    def init_matrices(self):
        'initialize the one-step basis and input effects matrices'

        dims = self.dims
        Timers.tic('expm')
        self.one_step_matrix_exp = expm(self.a_csc * self.time_elapser.step_size)
        Timers.toc('expm')

        Timers.tic('toarray')
        self.one_step_matrix_exp = self.one_step_matrix_exp.toarray()
        Timers.toc('toarray')

        if self.b_csc is not None:
            self.one_step_input_effects_matrix = np.zeros(self.b_csc.shape, dtype=float)

            for c in range(self.time_elapser.inputs):
                # create the a_matrix augmented with a column of the b_matrix as an affine term
                indptr = self.b_csc.indptr

                data = np.concatenate((self.a_csc.data, self. b_csc.data[indptr[c]:indptr[c+1]]))
                indices = np.concatenate((self.a_csc.indices, self.b_csc.indices[indptr[c]:indptr[c+1]]))
                indptr = np.concatenate((self.a_csc.indptr, [len(data)]))

                aug_a_csc = csc_matrix((data, indices, indptr), shape=(dims + 1, dims + 1))

                mat = aug_a_csc * self.time_elapser.step_size

                # the last column of matrix_exp is the same as multiplying it by the initial state [0, 0, ..., 1]
                init_state = np.zeros(dims + 1, dtype=float)
                init_state[dims] = 1.0
                col = expm_multiply(mat, init_state)

                self.one_step_input_effects_matrix[:, c] = col[:dims]

    def assign_basis_matrix(self, step_num):
        'first step matrix exp, other steps matrix multiplication'

        Timers.tic('init_matrices')
        if self.one_step_matrix_exp is None:
            self.init_matrices()
        Timers.toc('init_matrices')

        if step_num == 0: # step zero, basis matrix is identity matrix
            self.cur_basis_matrix = np.identity(self.dims, dtype=float)
            self.cur_input_effects_matrix = None
        elif step_num == 1:
            self.cur_basis_matrix = self.one_step_matrix_exp
            self.cur_input_effects_matrix = self.one_step_input_effects_matrix

            if self.use_lgg and self.b_csc is not None:
                prev_step_mat_exp = np.identity(self.dims, dtype=float)
                # make new (wider) input effects matrix
                blocks = [self.cur_input_effects_matrix, prev_step_mat_exp]
                self.cur_input_effects_matrix = np.concatenate(blocks, axis=1)
            
        elif step_num == self.cur_step + 1:
            Timers.tic('quick_step')
            prev_step_mat_exp = self.cur_basis_matrix
            self.cur_basis_matrix = np.dot(self.cur_basis_matrix, self.one_step_matrix_exp)

            # inputs
            if self.b_csc is not None:
                if self.use_lgg:
                    # cut cur_input_effects matrix into the relevant portion
                    self.cur_input_effects_matrix = self.cur_input_effects_matrix[:, 0:self.time_elapser.inputs]
                
                self.cur_input_effects_matrix = np.dot(self.one_step_matrix_exp, self.cur_input_effects_matrix)

                if self.use_lgg:
                    # make new (wider) input effects matrix
                    blocks = [self.cur_input_effects_matrix, prev_step_mat_exp]
                    self.cur_input_effects_matrix = np.concatenate(blocks, axis=1)

            Timers.toc('quick_step')
        else:
            Timers.tic('slow_step')

            Timers.tic('expm')
            # compute one step behind, because this is what's used by input effects matrix
            prev_step_mat_exp = expm(self.a_csc * (step_num-1) * self.time_elapser.step_size)
            Timers.toc('expm')

            # advance one step to get current basis matrix
            self.cur_basis_matrix = np.dot(prev_step_mat_exp.toarray(), self.one_step_matrix_exp)

            # inputs
            if self.b_csc is not None:
                Timers.tic("input effects")
                self.cur_input_effects_matrix = prev_step_mat_exp * self.one_step_input_effects_matrix
                Timers.toc("input effects")

                if self.use_lgg:
                    # make new (wider) input effects matrix
                    blocks = [self.cur_input_effects_matrix, prev_step_mat_exp]
                    self.cur_input_effects_matrix = np.concatenate(blocks, axis=1)

            Timers.toc('slow_step')

        self.cur_step = step_num

    def use_lgg_approx(self):
        '''
        set this time elapse object to use lgg approximation model
        '''

        self.use_lgg = True
        self.one_step_input_effects_matrix = self.b_csc.toarray() * self.time_elapser.step_size



'''
Hylaa Symoblic dynamics construction

Construct A matrix and reset matrix / rhs from symbolic expressions

Stanley Bak
Nov 2018
'''

import sympy
from sympy.parsing.sympy_parser import parse_expr
from sympy import Mul, Expr, Add, Symbol, Number

def extract_linear_terms(e, variables, has_affine_variable):
    '''extract linear terms from a flat sympy expression

    returns a list of numbers, one for each variable
    '''

    rv = [0] * len(variables)

    if has_affine_variable:
        rv.append(0)

    if not isinstance(e, Expr):
        raise RuntimeError("Expected sympy Expr: " + repr(e))

    try:
        _extract_linear_terms_rec(e, variables, rv, has_affine_variable)
    except RuntimeError as ex:
        raise RuntimeError(str(ex) + ", while parsing " + str(e))

    return rv

def _extract_linear_terms_rec(e, variables, rv, has_affine_variable):
    'extract linear terms'

    if isinstance(e, Add):
        _extract_linear_terms_rec(e.args[0], variables, rv, has_affine_variable)

        for arg in e.args[1:]:
            _extract_linear_terms_rec(arg, variables, rv, has_affine_variable)
    elif isinstance(e, Number):
        val = float(e)

        if val != 0:
            if not has_affine_variable:
                raise RuntimeError(f"expression has affine variables but has_affine_variable was False: '{e}'")

            rv[-1] += val
    elif isinstance(e, Symbol):
        try:
            index = variables.index(e.name)
        except ValueError:
            raise RuntimeError(f"variable {e.name} not found in variable list: {variables}")

        rv[index] += 1      
    elif isinstance(e, Mul):
        if len(e.args) != 2:
            raise RuntimeError(f"expected multiplication term with exactly two arguments: '{e}'")

        sym_term = None
        num_term = None

        if isinstance(e.args[0], Number) and isinstance(e.args[1], Symbol):
            num_term = e.args[0]
            sym_term = e.args[1]
        elif isinstance(e.args[1], Number) and isinstance(e.args[0], Symbol):
            num_term = e.args[1]
            sym_term = e.args[0]
        else:
            raise RuntimeError(f"expected multiplication with one number and one variable: '{e}'")

        try:
            index = variables.index(sym_term.name)
        except ValueError:
            raise RuntimeError(f"variable {sym_term.name} not found in variable list: {variables}")

        rv[index] += float(num_term)
    else:
        raise RuntimeError(f"unsupported term of type {type(e)}: '{e}'")

def make_reset_mat(variables, resets, constant_dict, has_affine_variable=False):
    'make the matrix for a reset operation'

    mat = make_dynamics_mat(variables, resets, constant_dict, has_affine_variable=has_affine_variable)

    if has_affine_variable:
        # affine variable should be identity reset, rather than all 0's as in dynamics
        mat[-1][-1] = 1.0

    return mat

def make_dynamics_mat(variables, derivatives, constant_dict, has_affine_variable=False):
    '''make the dynamics A matrix from the list of variables, derivatives, and a dict mapping constants to values

    returns a list of lists (a matrix) of size len(variables) by len(variables)
    '''

    rv = []
    subs = {}
    symbol_dict = {}
    
    for var in variables:
        symbol_dict[var] = sympy.symbols(var)
    
    for var, value in constant_dict.items():
        sym_var = sympy.symbols(var)
        subs[sym_var] = value
        symbol_dict[var] = sym_var

    for der in derivatives:
        sym_der = parse_expr(der, local_dict=symbol_dict)

        sym_der = sym_der.subs(subs)

        row = extract_linear_terms(sym_der, variables, has_affine_variable)
        rv.append(row)

    if has_affine_variable:
        rv.append([0] * (len(variables) + 1))

    return rv

def make_condition(variables, condition_list, constant_dict, has_affine_variable=False):
    '''make a condition matrix and right-hand-side (rhs) from a set of string conditions.
    condition_list is a list of strings with a single '<=' or '>=' condition like 'x - 1 + y <= 2 * x + 3'

    returns a 2-tuple: (mat, rhs)
    '''

    assert isinstance(condition_list, list), "condition_list should be a list of string conditions"

    mat = []
    rhs = []
    subs = {}
    symbol_dict = {}

    for var in variables:
        symbol_dict[var] = sympy.symbols(var)
    
    for var, value in constant_dict.items():
        sym_var = sympy.symbols(var)
        subs[sym_var] = value
        symbol_dict[var] = sym_var

    for cond in condition_list:
        less_than_count = cond.count('<=')
        greater_than_count = cond.count('>=')

        if less_than_count + greater_than_count != 1:
            raise RuntimeError(f"Expected condition with single '<=' or '>=': {cond}")

        if greater_than_count == 1:
            cond = cond.replace(">=", "<=")

        left, right = cond.split("<=")

        # swap left and right for '>=' expressions
        if greater_than_count == 1:
            left, right = right, left

        # make the expression: left - (right) <= 0
        subtract_cond = f"{left} - ({right})"
        
        sym_cond = parse_expr(subtract_cond, local_dict=symbol_dict)

        # substitute in constants
        sym_cond = sym_cond.subs(subs)

        terms = extract_linear_terms(sym_cond, variables, has_affine_variable=True)
        row = terms[:-1]

        if has_affine_variable:
            row.append(0)
        
        mat.append(row)
        rhs.append(-1 * terms[-1])

    return mat, rhs



'''
LP Utilities

This assumes a common LP structure, where the
first N columns correspond to the current-time variables, and
the first N rows are the current-time constraints (equality constraints equal to zero)
'''

import math

import numpy as np
import scipy as sp
from scipy.sparse import csr_matrix, csc_matrix, hstack

import swiglpk as glpk
from hylaa.lpinstance import LpInstance, SwigArray
from hylaa.timerutil import Timers

def from_box(box_list, mode):
    'make a new lp instance from a passed-in box'

    rhs = []

    for lb, ub in box_list:
        assert lb <= ub, "lower bound ({}) > upper bound ({})".format(lb, ub)
        rhs.append(-lb)
        rhs.append(ub)

    # make constraints as csr_matrix
    dims = len(box_list)
    data = []
    inds = []
    indptr = [0]

    # -1 <= -lb
    # 1 <= ub
    for n in range(dims):
        data.append(-1)
        inds.append(n)
        indptr.append(len(data))

        data.append(1)
        inds.append(n)
        indptr.append(len(data))

    csr = csr_matrix((data, inds, indptr), shape=(2*dims, dims), dtype=float)
    csr.check_format()

    return from_constraints(csr, rhs, mode)

def from_zonotope(center, generator_list, mode):
    'make a new lp instance from the passed in zonotope'

    #variables [x1, x2, ..., center_var, alpha1, alpha2, ...]
    mat = []
    rhs = []

    cdims = len(center)
    dims = cdims + 1 + len(generator_list)

    # bounds on center_var (center_var = 1)
    mat.append([1 if d == cdims else 0 for d in range(dims)])
    rhs.append(1)

    mat.append([-1 if d == cdims else 0 for d in range(dims)])
    rhs.append(-1)

    for i, g in enumerate(generator_list):
        assert len(g) == cdims, "expected each generator to have the same number of dims as the center"
        
        # bounds on each alpha (-1 <= alpha <= 1)
        mat.append([1 if d == cdims + 1 + i else 0 for d in range(dims)])
        rhs.append(1)

        mat.append([-1 if d == cdims + 1 + i else 0 for d in range(dims)])
        rhs.append(1)

    # zonotope generator constraints x = c + alpha1 * g1 + alpha2 * g2 + ...
    for dim, c in enumerate(center):
        row = [0] * dims

        row[dim] = -1
        row[cdims] = c

        for gindex, generator in enumerate(generator_list):     
            row[cdims + 1 + gindex] = generator[dim]

        mat.append(row)
        rhs.append(0)

        mat.append([-1 * x for x in row])
        rhs.append(0)

    return from_constraints(mat, rhs, mode, dims=cdims)

def from_constraints(csr, rhs, mode, types=None, names=None, dims=None):
    '''make a new lp instance from a passed-in set of constraints and rhs

    if types/names is None, then assume all constraints are '<=' constriants

    if dims is None, assume the number of columns in the csr matrix is the number of variables
    otherwise, assume the left-most columns in the csr_matrix are the current-time variables 

    '''

    if not isinstance(csr, csr_matrix):
        csr = csr_matrix(csr, dtype=float)

    if not isinstance(rhs, np.ndarray):
        rhs = np.array(rhs, dtype=float)

    assert len(rhs.shape) == 1
    assert csr.shape[0] == len(rhs)
    assert is_feasible(csr, rhs), "initial constraints are not feasible"
    assert (types is None) == (names is None)

    if dims is None:
        dims = csr.shape[1]
    else:
        assert dims <= csr.shape[1]

    lpi = LpInstance()
    lpi.add_rows_equal_zero(dims)

    if names is None:
        names = ["m{}_i{}".format(mode.mode_id, var_index) for var_index in range(csr.shape[1])]
        
    names += ["m{}_c{}".format(mode.mode_id, var_index) for var_index in range(dims)]
    
    lpi.add_cols(names)

    if types is None:
        lpi.add_rows_less_equal(rhs)
    else:
        assert len(types) == len(rhs)
        lpi.add_rows_with_types(types, rhs)

    has_inputs = mode.b_csr is not None

    if has_inputs:
        names = ["m{}_ti{}".format(mode.mode_id, n) for n in range(dims)]
        lpi.add_cols(names)

    # make constraints as csr_matrix
    data = []
    inds = []
    indptr = [0]

    # I 0 -I for first n rows
    for n in range(dims):
        data.append(1)
        inds.append(n)

        data.append(-1)
        inds.append(csr.shape[1] + n)

        if has_inputs:
            data.append(1)
            inds.append(csr.shape[1] + dims + n)

        indptr.append(len(data))
        
    num_cols = csr.shape[1] + dims if not has_inputs else 2*dims + csr.shape[1]
    basis_constraints = csr_matrix((data, inds, indptr), shape=(dims, num_cols), dtype=float)
    basis_constraints.check_format()

    lpi.set_constraints_csr(basis_constraints)

    # add constraints on initial conditions, offset by <dims> rows
    lpi.set_constraints_csr(csr, offset=(dims, 0))

    # add total input effects rows
    if has_inputs:
        rows_before = lpi.get_num_rows()
        ie_pos = (rows_before, dims + csr.shape[1])
        lpi.add_rows_equal_zero(dims)

        # -I
        csr_ti = -1 * sp.sparse.identity(dims, dtype=float, format='csr')
        lpi.set_constraints_csr(csr_ti, offset=ie_pos)
    else:
        ie_pos = None

    cur_vars_offset = csr.shape[1]
    lpi.set_reach_vars(dims, (0, 0), cur_vars_offset, ie_pos)

    return lpi

def set_basis_matrix(lpi, basis_mat):
    'modify the lpi in place to set the basis matrix'

    assert basis_mat.shape[0] == basis_mat.shape[1], "expected square matrix"
    assert basis_mat.shape[0] == lpi.dims, \
      f"basis matrix wrong shape, expected ({lpi.dims}, {lpi.dims}), got {basis_mat.shape}"

    # this is done using the optimized swigvec interface in lpinstance

    data_vec_list = [] # list of swig doubleArray objects for each row

    # 0 BM 0 -I 0 (I? <- if inputs exist)
    for row in range(lpi.dims):
        data = []

        data += basis_mat[row].tolist()
        
        data.append(-1.0)

        if lpi.input_effects_offsets is not None:
            data.append(1.0)

        data_vec_list.append(SwigArray.as_double_array(data))

    entries_per_row = basis_mat.shape[0] + 1 + (1 if lpi.input_effects_offsets else 0)
    count_list = [entries_per_row] * basis_mat.shape[0]
        
    lpi.set_constraints_swigvec_rows(data_vec_list, lpi.bm_indices, count_list, lpi.basis_mat_pos[0])

def add_input_effects_matrix(lpi, input_mat, mode, lgg_beta=None):
    'add an input effects matrix to this lpi'

    assert lpi.input_effects_offsets is not None
    assert mode.b_csr is not None

    num_vars = mode.a_csr.shape[1]
    num_inputs = mode.b_csr.shape[1]
    num_constraints = len(mode.u_constraints_rhs)

    # if lgg approximation model is used, then input effects will also bloat in every dimension by beta
    # this makes the input effects matrix wider
    if lgg_beta is None:
        assert input_mat.shape[1] == num_inputs
    else:
        assert input_mat.shape[1] + num_inputs + num_vars
    
    assert input_mat.shape[0] == num_vars
    assert lpi.dims == num_vars

    # add new row/cols
    names = ["m{}_I{}".format(mode.mode_id, i) for i in range(num_inputs)]

    if lgg_beta is not None: # bloating variables
        names += ["m{}_b{}".format(mode.mode_id, i) for i in range(num_vars)]
    
    pre_cols = lpi.get_num_cols()
    lpi.add_cols(names)

    pre_rows = lpi.get_num_rows()
    lpi.add_rows_less_equal(mode.u_constraints_rhs)

    if lgg_beta is not None: # bloating constraints
        pre_bloat_rows = lpi.get_num_rows()
        lpi.add_rows_less_equal([lgg_beta] * (2 * num_vars))

    # set constraints on the rows/cols, as well as the input basis matrix using a csc_matrix
    data = []
    inds = []
    indptr = [0]

    for c in range(num_inputs):
        # input basis matrix column c
        for row in range(lpi.dims):
            data.append(input_mat[row, c])
            inds.append(row + lpi.input_effects_offsets[0])

        # constraints_csr column c
        for i in range(mode.u_constraints_csc.indptr[c], mode.u_constraints_csc.indptr[c+1]):
            row = mode.u_constraints_csc.indices[i]
            val = mode.u_constraints_csc.data[i]

            data.append(val)
            inds.append(pre_rows + row)

        indptr.append(len(data))

    if lgg_beta is not None: # constraints for bloating
        for v in range(num_vars):
            # input basis matrix column num_inputs + v
            for row in range(num_vars):
                data.append(input_mat[row, num_inputs + v])
                inds.append(row + lpi.input_effects_offsets[0])

            # constraints_csr column corresponding to variable v
            data.append(1)
            inds.append(pre_bloat_rows + 2*v)

            data.append(-1)
            inds.append(pre_bloat_rows + 2*v + 1)

            indptr.append(len(data))

    num_cols = num_inputs
    num_rows = pre_rows + num_constraints

    if lgg_beta is not None:
        num_rows += 2 * num_vars
        num_cols += num_vars
    
    csc = csc_matrix((data, inds, indptr), shape=(num_rows, num_cols), dtype=float)
    csc.check_format()

    lpi.set_constraints_csc(csc, offset=(0, pre_cols))

def check_intersection(lpi, lc, tol=1e-13):
    '''check if there is an intersection between the LP constriants and the LinearConstraint object lc

    This solves an LP optimizing in the given direction... without adding the constraint to the LP

    This returns True/False if an intersection is possible

    it also can return None if the lp is infeasible
    '''

    Timers.tic("check_intersection")

    lpi.set_minimize_direction(lc.csr, is_csr=True)

    columns = lc.csr.indices[0:lc.csr.indptr[1]]
    lp_columns = [lpi.cur_vars_offset + c for c in columns]

    lp_res = lpi.minimize(columns=lp_columns, fail_on_unsat=False)

    if lp_res is None:
        # sometimes, changing optimization direction makes lp infeasible (up to numerical accuracy)
        # this happens in gearbox with small time steps. In this case, return no intersection
        rv = None
    else:
        dot_res = np.dot(lc.csr.data, lp_res)
        rv = dot_res + tol <= lc.rhs

    Timers.toc("check_intersection")

    return rv

def add_init_constraint(lpi, vec, rhs, basis_matrix=None, input_effects_list=None, row_index=None):
    '''
    add a constraint to the lpi

    this replaces an existing row or adds a new one (if row_index is None), with constraints assigned to the 
    initial-time variables

    this returns the row of the newly-created constraint
    '''

    if basis_matrix is None:
        basis_matrix = get_basis_matrix(lpi)

    assert isinstance(basis_matrix, np.ndarray)

    # we need to project the basis matrix using the passed in direction vector
    preshape = vec.shape

    dims = basis_matrix.shape[0]
    vec.shape = (1, dims) # vec is now the projection matrix for this direction
    bm_projection = np.dot(vec, basis_matrix)

    rows = lpi.get_num_rows()
    cols = lpi.get_num_cols()

    if row_index is None:
        row_index = rows
        lpi.add_rows_less_equal([rhs])
    else:
        # overwrite the old rhs
        lpi.set_constraint_rhs(row_index, rhs)

    indptr = [0]

    # basis matrix
    inds = [lpi.basis_mat_pos[1] + i for i in range(dims)]
    data = [val for val in bm_projection[0]]

    # each of the input effects matrices
    if input_effects_list:
        num_inputs = input_effects_list[0].shape[1]
        offset = lpi.input_effects_offsets[1] + dims

        for ie_mat in input_effects_list:
            ie_projection = np.dot(vec, ie_mat)[0]

            assert len(ie_projection) == num_inputs, \
              f"len(ie_projection) ({len(ie_projection)}) != num_inputs ({num_inputs})"

            inds += [offset + i for i in range(num_inputs)]
            data += [val for val in ie_projection]
            offset += num_inputs

    indptr.append(len(data))

    # restore vector shape to what it was when passed in
    vec.shape = preshape

    csr_row_mat = csr_matrix((data, inds, indptr), dtype=float, shape=(1, cols))
    csr_row_mat.check_format()

    lpi.set_constraints_csr(csr_row_mat, offset=(row_index, 0))

    return row_index

def try_replace_init_constraint(lpi, old_row_index, direction, rhs, basis_mat=None, input_effects_list=None):
    '''replace the constraint in row_index by a new constraint, if the new constraint is stronger, otherwise
    create new constriant

    this is used for removing redundant invariant constraints

    This returns (row_index, is_stronger) where row_index is the index of the new constraint and is_stronger is
    a boolean indicating if the old constraint was completely replaced (the new constraint was strictly stronger)
    '''

    is_stronger = True

    if basis_mat is None:
        basis_mat = get_basis_matrix(lpi)

    # Improved algorithm to do this: Replace the constraint, then optimize in the direction of the old constraint to
    # check if the old constraint is still feasible.
    # this avoids deleting rows and therefore possibly ruining warm start if the constraint was non-basic

    # first extract the old constraint
    old_constraint = lpi.get_row(old_row_index)
    old_rhs = lpi.get_rhs([old_row_index])[0]
        
    # replace the old constraint with the new one
    new_row_index = add_init_constraint(lpi, direction, rhs, basis_matrix=basis_mat, \
                                        input_effects_list=input_effects_list, row_index=old_row_index)

    # optimize in the direction of the old constraint, to see if the old constraint is still feasible
    lpi.set_minimize_direction(-1 * old_constraint, is_csr=True, offset=0)

    res = lpi.minimize(fail_on_unsat=False)

    # if the lp was unsat, this means adding the new constraint makes it infeasible, so it's safe to replace
    if res is not None:
        res_dot = (old_constraint * np.array(res, dtype=float))

        res_dot = res_dot[0]

        # if res_dot < old_rhs, then the old constraint is no longer needed, otherwise, re-add it
        if res_dot >= old_rhs:
            lpi.add_rows_less_equal([old_rhs])
            rows = lpi.get_num_rows()

            lpi.set_constraints_csr(old_constraint, offset=(rows - 1, 0))
            is_stronger = False

    return new_row_index, is_stronger

def aggregate_chull(lpi_list, mode):
    '''
    perform aggregation using convex hull (non-recursive call)

    This uses the sop closed convex hull algorithm from Willem Hagemann. See his PhD Dissertation:
    "Symbolic Orthogonal Projections: A New Polyhedral Representation for Reachability Analysis of Hybrid Systems"
    '''

    if len(lpi_list) == 1:
        return lpi_list[0].clone()

    # make all the base LP's complete by adding the redudant row 0 * x <= 1
    new_lpi_list = []

    for old_lpi in lpi_list:
        lpi = old_lpi.clone()
        lpi.add_rows_less_equal([1])
        new_lpi_list.append(lpi)

    lpi = _aggregate_chull_recursive(new_lpi_list, mode)

    # need to essentially take snapshot variables
    csr = lpi.get_full_constraints()
    rhs = lpi.get_rhs()
    types = lpi.get_types()
    names = lpi.get_names()

    # from_constraints assumes left-most variables are current-time variables
    return from_constraints(csr, rhs, mode, types=types, names=names, dims=lpi.dims)

def _aggregate_chull_recursive(lpi_list, mode):
    '''
    perform aggregation using convex hull (recursive call)

    This uses the sop closed convex hull algorithm from Willem Hagemann. See his PhD Dissertation:
    "Symbolic Orthogonal Projections: A New Polyhedral Representation for Reachability Analysis of Hybrid Systems"
    '''

    assert len(lpi_list) >= 2

    # recursive cases:
    #if isinstance(lpi_list, LpInstance):
    #    return lpi_list

    if len(lpi_list) > 2:
        mid = len(lpi_list) // 2

        return aggregate_chull([aggregate_chull(lpi_list[:mid], mode), aggregate_chull(lpi_list[mid:], mode)], mode)

    # base case: exactly two lpis in lpi_list
    assert len(lpi_list) == 2

    dims = lpi_list[0].dims
    assert lpi_list[1].dims == dims

    lpi1 = lpi_list[0]
    csr1 = lpi1.get_full_constraints()
    rhs1 = lpi1.get_rhs()
    types1 = lpi1.get_types()

    # csr1 contains L_left A L_right, split it into these three in order to construct L
    l_left = csr1[:, 0:lpi1.cur_vars_offset]
    a1 = csr1[:, lpi1.cur_vars_offset:lpi1.cur_vars_offset+dims]
    l_right = csr1[:, lpi1.cur_vars_offset+dims:]

    l1 = hstack([l_left, l_right])

    # repeat for lpi2
    lpi2 = lpi_list[1]
    csr2 = lpi2.get_full_constraints()
    rhs2 = lpi2.get_rhs()
    types2 = lpi2.get_types()

    # csr1 contains L_left A L_right, split it into these three in order to construct L
    l_left = csr2[:, 0:lpi2.cur_vars_offset]
    a2 = csr2[:, lpi2.cur_vars_offset:lpi2.cur_vars_offset+dims]
    l_right = csr2[:, lpi2.cur_vars_offset+dims:]
    l2 = hstack([l_left, l_right])

    lpi = LpInstance()
    #lpi.add_rows_equal_zero(dims)

    types = types1 + types2
    rhs = [n for n in rhs1] + ([0] * len(rhs2))

    lpi.add_rows_with_types(types, rhs)

    cols = []

    cols += [f"A1_{i}" for i in range(dims)]
    cols += [f"A2_{i}" for i in range(dims)]
    cols += [f"L1_{i}" for i in range(l1.shape[1])]
    cols += [f"L2_{i}" for i in range(l2.shape[1])]
    cols += ["a"]

    lpi.add_cols(cols)

    # set constraints
    l2_zero = csr_matrix((a1.shape[0], l2.shape[1])) # the 0 above L2
    l1_zero = csr_matrix((a2.shape[0], l1.shape[1])) # the 0 below L1
    a1_zero = csr_matrix((a2.shape[0], a1.shape[1])) # the 0 below A1

    rhs1_vmat = csr_matrix(np.array([[num] for num in rhs1]))
    rhs2_vmat = csr_matrix(np.array([[num] for num in rhs2]))

    top = csr_matrix(hstack([a1, a1, l1, l2_zero, rhs1_vmat]))
    lpi.set_constraints_csr(top)

    bottom = csr_matrix(hstack([a1_zero, -a2, l1_zero, l2, -rhs2_vmat]))
    lpi.set_constraints_csr(bottom, offset=(top.shape[0], 0))

    lpi.dims = dims
    lpi.cur_vars_offset = 0 # left-most variables are current time variables

    return lpi

def aggregate(lpi_list, direction_matrix, mode):
    '''
    return a new lpi consisting of an aggregation of the passed-in lpi list

    This creates a template polytope using the passed-in directions (passed in as rows of direction_matrix).

    use lputil.make_direction_matrix() to create the direction_matrix with arnoldi directions
    '''

    assert isinstance(direction_matrix, np.ndarray)
    assert direction_matrix.dtype == float
    assert direction_matrix.shape[0] >= direction_matrix.shape[1], "expected num directions >= dims"
    assert len(lpi_list) > 1, "expected more than one lpi to perform an aggregation"

    inds = []
    data = []
    indptrs = [0]
    rhs = []

    # for each direction, minimize and maximize it within the list
    for direction in direction_matrix:
        if abs(np.linalg.norm(direction)) < 1e-6:
            continue
        #assert abs(np.linalg.norm(direction) - 1) < 1e-9, "expected normalized directions, got {}".format(direction)

        dir_inds = [i for i, x in enumerate(direction) if x != 0]
        dir_data = [x for x in direction if x != 0]
        dir_neg_data = [-x for x in dir_data]

        max_val = -np.inf
        min_val = np.inf
       
        for lpi in lpi_list:
            assert direction_matrix.shape[1] == lpi.dims

            dir_columns = [i + lpi.cur_vars_offset for i in dir_inds]

            result = lpi.minimize(direction_vec=-direction, columns=dir_columns)
            max_val = max(max_val, np.dot(result, dir_data))
            
            result = lpi.minimize(direction_vec=direction, columns=dir_columns)
            min_val = min(min_val, np.dot(result, dir_data))

        inds += dir_inds
        data += dir_data
        indptrs.append(len(data))
        rhs.append(max_val)

        inds += dir_inds
        data += dir_neg_data
        indptrs.append(len(data))
        rhs.append(-min_val)

    rows = len(indptrs) - 1
    cols = direction_matrix.shape[1]
    csr_mat = csr_matrix((data, inds, indptrs), dtype=float, shape=(rows, cols))
    csr_mat.check_format()
    
    rv = from_constraints(csr_mat, rhs, mode)

    return rv

def get_basis_matrix(lpi):
    'get the basis matrix from the lpi'

    row = lpi.basis_mat_pos[0]
    col = lpi.basis_mat_pos[1]

    # get_dense_constraints takes in x and y
    return lpi.get_dense_constraints(col, row, lpi.dims, lpi.dims)

def scale_with_bm(lpi, amount):
    '''
    scale the current set using the basis matrix
    '''

    bm = get_basis_matrix(lpi)
    set_basis_matrix(lpi, amount * bm)

def bloat(lpi, amount, var_name_prefix='bloat'):
    '''
    bloat the current set of states
    '''

    assert amount >= 0

    # strategy add n variables with bounds -amount <= x <= amount
    names = [f"{var_name_prefix}{n}" for n in range(lpi.dims)]
    precols = lpi.get_num_cols()
    prerows = lpi.get_num_rows()

    data = []
    inds = []
    indptr = [0]

    for n in range(lpi.dims):
        # 1.0 entry in the basis matrix row for variable n
        data.append(1)
        inds.append(lpi.basis_mat_pos[0] + n)
        
        # x <= amount
        data.append(1)
        inds.append(prerows + 2*n)

        # -x <= amount ---> x >= -amount
        data.append(-1)
        inds.append(prerows + 2*n + 1)
        indptr.append(len(data))

    mat = csc_matrix((data, inds, indptr), shape=(prerows + 2*lpi.dims, lpi.dims), dtype=float)

    rhs = [amount] * (2 * lpi.dims)
    lpi.add_rows_less_equal(rhs)
    lpi.add_cols(names)

    lpi.set_constraints_csc(mat, offset=(0, precols))

def add_reset_variables(lpi, mode_id, transition_index, # pylint: disable=too-many-locals, too-many-statements
                        reset_csr=None, minkowski_csr=None,
                        minkowski_constraints_csr=None, minkowski_constraints_rhs=None, successor_has_inputs=False): 
    '''
    add variables associated with a reset

    general resets are of the form x' = Rx + My, Cy <= rhs, where y are fresh variables
    the reset_minkowski variables can be None if no new variables are needed. If unassigned, the identity
    reset is assumed

    x' are the new variables
    x are the old variables       
    reset_csr is R (None -> identity)
    minkowski_csr is M
    minkowski_constraints_csr is C
    minkowski_constraints_rhs is rhs

    this function adds new variables for both the initial states and the current states in the new mode
    '''

    old_dims = lpi.dims
    cols = lpi.get_num_cols()
    rows = lpi.get_num_rows()

    if reset_csr is None:
        reset_csr = sp.sparse.identity(old_dims, dtype=float, format='csr')

    if minkowski_csr is None:
        minkowski_csr = csr_matrix((0, 0))
        minkowski_constraints_csr = csr_matrix((0, 0))
        minkowski_constraints_rhs = np.array([])

    assert isinstance(reset_csr, csr_matrix)
    assert isinstance(minkowski_csr, csr_matrix)
    assert old_dims == reset_csr.shape[1], "Reset matrix shape is wrong (expected {} cols)".format(old_dims)

    # it may be possible to change the number of dimensions between modes
    new_dims = reset_csr.shape[0]

    min_vars = minkowski_csr.shape[1]

    names = ["reset{}".format(min_var) for min_var in range(min_vars)]
    names += ["m{}_i0_t{}".format(mode_id, transition_index)]

    names += ["m{}_i{}".format(mode_id, d) for d in range(1, new_dims)]
    names += ["m{}_c{}".format(mode_id, d) for d in range(new_dims)]

    if successor_has_inputs:
        names += ["m{}_ti{}".format(mode_id, d) for d in range(new_dims)]
    
    lpi.add_cols(names)

    lpi.add_rows_equal_zero(2*new_dims)

    lpi.add_rows_less_equal(minkowski_constraints_rhs)

    if successor_has_inputs:
        lpi.add_rows_equal_zero(new_dims)

    data = []
    inds = []
    indptrs = [0]

    # new_init_vars = reset_mat * old_cur_vars + minkow_csr * minkow_vars:
    # -I for new mode initial vars, RM for old mode cur_vars, MK for minkow_vars
    for dim in range(new_dims):

        # old cur_vars
        for index in range(reset_csr.indptr[dim], reset_csr.indptr[dim + 1]):
            rm_col = reset_csr.indices[index]
            value = reset_csr.data[index]
            
            data.append(value)
            inds.append(lpi.cur_vars_offset + rm_col)

        # minkow_vars
        if minkowski_csr.shape[1] > 0:
            for index in range(minkowski_csr.indptr[dim], minkowski_csr.indptr[dim + 1]):
                minkowski_col = minkowski_csr.indices[index]
                value = minkowski_csr.data[index]

                data.append(value)
                inds.append(cols + minkowski_col)

        # new mode initial vars
        data.append(-1)
        inds.append(cols + min_vars + dim)

        indptrs.append(len(data))

    # new_cur_vars = BM * new_init_vars: -I for new cur vars, BM (initially identity) for new init vars
    for dim in range(new_dims):
        data.append(1)
        inds.append(cols + min_vars + dim)

        data.append(-1)
        inds.append(cols + min_vars + new_dims + dim)

        if successor_has_inputs:
            data.append(1)
            inds.append(cols + min_vars + 2*new_dims + dim)

        indptrs.append(len(data))

    # encode minkowski constraint rows
    for row in range(minkowski_constraints_csr.shape[0]):
        for index in range(minkowski_constraints_csr.indptr[row], minkowski_constraints_csr.indptr[row + 1]):
            col = minkowski_constraints_csr.indices[index]
            value = minkowski_constraints_csr.data[index]
            
            data.append(value)
            inds.append(cols + col)

        indptrs.append(len(data))

    # encode total input effects
    if successor_has_inputs:
        for dim in range(new_dims):
            data.append(-1)
            inds.append(cols + min_vars + 2*new_dims + dim)

            indptrs.append(len(data))

    height = 2*new_dims + len(minkowski_constraints_rhs)
    width = cols + 2*new_dims + min_vars

    if successor_has_inputs:
        height += new_dims
        width += new_dims

    mat = csr_matrix((data, inds, indptrs), dtype=float, \
                     shape=(height, width))
    mat.check_format()

    lpi.set_constraints_csr(mat, offset=(rows, 0))

    # input effects position
    if successor_has_inputs:
        ie_x = cols + min_vars + 2*new_dims
        ie_y = rows + 2*new_dims + len(minkowski_constraints_rhs)
        ie_offsets = (ie_y, ie_x)
    else:
        ie_offsets = None

    basis_mat_pos = (rows+new_dims, cols + minkowski_csr.shape[1])
    cur_vars_offset = cols + minkowski_csr.shape[1] + new_dims
    
    lpi.set_reach_vars(new_dims, basis_mat_pos, cur_vars_offset, ie_offsets)

def add_curtime_constraints(lpi, csr, rhs_vec):
    '''
    add constraints to the lpi

    this adds them on the current time variables (not the initial time variables)
    '''

    assert isinstance(csr, csr_matrix)

    prerows = lpi.get_num_rows()
    lpi.add_rows_less_equal(rhs_vec)

    lpi.set_constraints_csr(csr, offset=(prerows, lpi.cur_vars_offset))

def get_box_center(lpi):
    '''get the center of the box overapproximation of the passed-in lpi

    may return None if lp solving fails (numerical issues)
    '''

    Timers.tic('get_box_center')

    dims = lpi.dims
    pt = []

    for dim in range(dims):
        col = lpi.cur_vars_offset + dim
        min_dir = [1 if i == dim else 0 for i in range(dims)]
        max_dir = [-1 if i == dim else 0 for i in range(dims)]
        
        min_val = lpi.minimize(direction_vec=min_dir, columns=[col], fail_on_unsat=False)
        max_val = lpi.minimize(direction_vec=max_dir, columns=[col], fail_on_unsat=False)

        if min_val is None or max_val is None:
            pt = None
            break
        
        min_val = min_val[0]
        max_val = max_val[0]

        pt.append((min_val + max_val) / 2.0)

    Timers.toc('get_box_center')

    return pt

def make_direction_matrix(point, a_csr):
    '''make the direction matrix for arnoldi aggregation

    this is a set of full rank, linearly-independent vectors, extracted from the dynamics using something
    similar to the arnoldi iteration

    the null-space vectors first try to be filled with the orthonormal directions, and then by random vectors

    point is the point where to sample the dynamics
    a_csr is the dynamics matrix
    '''

    Timers.tic('make_direction_matrix')

    assert isinstance(a_csr, csr_matrix)
    cur_vec = np.array(point, dtype=float)
    
    assert len(point) == a_csr.shape[1], "expected point dims({}) to equal A-matrix dims({})".format( \
                len(point), a_csr.shape[1])
    
    dims = len(point)
    rv = []

    while len(rv) < dims:
        if cur_vec is None: # inside the null space
            # first try to pick orthonormal directions if we can
            for d in range(dims):
                found_nonzero = False

                for vec in rv:
                    if vec[d] != 0:
                        found_nonzero = True
                        break

                if found_nonzero is False:
                    cur_vec = np.array([1 if n == d else 0 for n in range(dims)], dtype=float)
                    break

            # if that didn't work, just a random vector
            if cur_vec is None:
                cur_vec = np.random.rand(dims,)
        else:
            cur_vec = a_csr * cur_vec

        # project out the previous vectors
        for prev_vec in rv:
            dot_val = np.dot(prev_vec, cur_vec)

            cur_vec -= prev_vec * dot_val

        norm = np.linalg.norm(cur_vec, 2)

        assert not math.isinf(norm) and not math.isnan(norm), "vector norm was infinite in arnoldi"

        if norm < 1e-6:
            # super small norm... basically it's in the subspace spanned by previous vectors, restart
            cur_vec = None
        else:
            cur_vec = cur_vec / norm

            rv.append(cur_vec)

    Timers.toc('make_direction_matrix')

    return np.array(rv, dtype=float)

def reorthogonalize_matrix(mat, dims):
    '''given an input matrix (one 'dims'-dimensional vector per row), return a new matrix such that the vectors are in 
    the same order, but orthonormal (project out earlier vectors and scale), with the passed-in number of dimensions 
    (a smaller matrix may be returned, or new vectors may be generated to fill the nullspace if the dims > dim(mat)'''

    if isinstance(mat, list):
        mat = np.array(mat, dtype=float)

    assert mat.shape[1] == dims, "mat should have width equal to dims({})".format(dims)

    # take approach similar to arnoldi, except without the matrix-vector multiplication (see make_direction_matrix)

    Timers.tic('reorthogonalize_matrix')

    rv = []

    next_index = 0

    while len(rv) < dims:
        if next_index >= mat.shape[0]:
            # first try to pick orthonormal directions if we can
            for d in range(dims):
                found_nonzero = False

                for vec in rv:
                    if vec[d] != 0:
                        found_nonzero = True
                        break

                if found_nonzero is False:
                    cur_vec = np.array([1 if n == d else 0 for n in range(dims)], dtype=float)
                    break

            # if that didn't work, just a random vector
            if cur_vec is None:
                cur_vec = np.random.rand(dims,)
        else:
            cur_vec = mat[next_index]
            next_index += 1

        # project out the previous vectors
        for prev_vec in rv:
            dot_val = np.dot(prev_vec, cur_vec)

            cur_vec -= prev_vec * dot_val

        norm = np.linalg.norm(cur_vec, 2)

        assert not math.isinf(norm) and not math.isnan(norm), "vector norm was infinite in arnoldi"

        if norm < 1e-6:
            # super small norm... basically it's in the subspace spanned by previous vectors, restart
            cur_vec = None
        else:
            cur_vec = cur_vec / norm

            rv.append(cur_vec)

    Timers.toc('reorthogonalize_matrix')

    return np.array(rv, dtype=float)

def is_feasible(csr, rhs):
    'are the passed in constraints feasible?'

    if not isinstance(csr, csr_matrix):
        csr = csr_matrix(csr, dtype=float)

    assert len(rhs) == csr.shape[0], "constraints RHS differs from number of rows"

    lpi = LpInstance()
    names = ["x{}".format(n) for n in range(csr.shape[1])]
    lpi.add_cols(names)
    lpi.add_rows_less_equal(rhs)

    lpi.set_constraints_csr(csr)

    lpi.cur_vars_offset = 0
    lpi.dims = 0
    lpi.basis_mat_pos = (0, 0)

    return lpi.is_feasible()

def is_point_in_lpi(point, orig_lpi):
    '''is the passed-in point in the lpi?

    This function is strictly for unit testing as it's slow (copies the lpi). 
    A warning is printed to stdout to reflect this and discourage usage in other places.
    '''

    print("Warning: Using testing function lputil.is_point_in_lpi (slow)")

    assert len(point) <= orig_lpi.dims

    inds = []
    data = []
    indptr = [0]
    rhs = []

    for i, x in enumerate(point):
        inds.append(i)
        data.append(1)
        indptr.append(len(data))
        rhs.append(x)

        inds.append(i)
        data.append(-1)
        indptr.append(len(data))
        rhs.append(-x)

    rows = len(indptr) - 1
    cols = len(point)
    csr = csr_matrix((data, inds, indptr), dtype=float, shape=(rows, cols))

    lpi = orig_lpi.clone()
    add_curtime_constraints(lpi, csr, rhs)

    return lpi.is_feasible()

def compute_radius_inf(lpi):
    '''
    compute the max ||x||, x in lpi, according to the infinity norm

    This uses 2*n LPs and then returns the maximum over all components
    '''

    max_val = -float('inf')

    for n in range(lpi.dims):
        for posneg in [1.0, -1.0]:
            dir_vec = [posneg if n == d else 0.0 for d in range(lpi.dims)]

            lp_result = lpi.minimize(dir_vec)

            max_val = max(max_val, abs(lp_result[n]))

    return max_val

def minkowski_sum(lpi_list, mode):
    '''
    perform a minkowski sum of the passed-in sets, and return the resultant lpi
    '''

    for lpi in lpi_list:
        assert lpi.dims == lpi_list[0].dims, "dimension mismatch during minkowski sum"

    dims = lpi_list[0].dims

    csr_list = []
    combined_rhs = [0] * dims
    combined_types = [glpk.GLP_FX] * dims
    combined_names = [f"c{n}" for n in range(dims)]

    total_new_vars = dims

    for i, lpi in enumerate(lpi_list):
        csr = lpi.get_full_constraints()
        csr_list.append(csr)
        combined_rhs += [v for v in lpi.get_rhs()]
        combined_types += lpi.get_types()

        total_new_vars += csr.shape[1]
        combined_names += [f"l{i}_{v}" for v in range(csr.shape[1])]

    # create combined_csr constraints
    data = []
    indices = []
    indptr = [0]

    for d in range(dims):
        data.append(1)
        indices.append(d)
        col_offset = dims

        for lpi in lpi_list:
            data.append(-1)
            indices.append(col_offset + lpi.cur_vars_offset + d)

            col_offset += lpi.get_num_cols()

        indptr.append(len(data))

    # copy constraints from each lpi
    col_offset = dims
    indptr_offset = indptr[-1]
    
    for csr in csr_list:
        data += [d for d in csr.data]
        indices += [col_offset + i for i in csr.indices]
        indptr += [indptr_offset + i for i in csr.indptr[1:]]

        col_offset += csr.shape[1]
        indptr_offset = indptr[-1]

    rows = len(combined_rhs)
    cols = col_offset
    combined_csr = csr_matrix((data, indices, indptr), shape=(rows, cols), dtype=float)

    # from_constraints assumes left-most variables are current-time variables
    return from_constraints(combined_csr, combined_rhs, mode, types=combined_types, names=combined_names, dims=dims)

def from_input_constraints(b_mat, u_constraints, u_rhs, mode):
    'create an lpi from input constraints (B matrix and constraints on U)'

    if not isinstance(b_mat, csr_matrix):
        b_mat = csr_matrix(b_mat, dtype=float)

    if not isinstance(u_constraints, csr_matrix):
        u_constraints = csr_matrix(u_constraints, dtype=float)

    if not isinstance(u_rhs, np.ndarray):
        u_rhs = np.array(u_rhs, dtype=float)

    # V = B U
    dims = b_mat.shape[0]

    combined_rhs = [0] * dims
    combined_types = [glpk.GLP_FX] * dims
    combined_names = [f"c{n}" for n in range(dims)]

    combined_rhs += [v for v in u_rhs]
    combined_types += [glpk.GLP_UP] * u_constraints.shape[0]
    combined_names += [f"u{n}" for n in range(b_mat.shape[1])]

    # create combined_csr constraints
    data = []
    indices = []
    indptr = [0]

    for d in range(dims):
        data.append(-1)
        indices.append(d)

        data += [d for d in b_mat.data[b_mat.indptr[d]:b_mat.indptr[d+1]]]
        indices += [dims + i for i in b_mat.indices[b_mat.indptr[d]:b_mat.indptr[d+1]]]

        indptr.append(len(data))

    indptr_offset = indptr[-1]
    
    data += [d for d in u_constraints.data]
    indices += [dims + i for i in u_constraints.indices]
    indptr += [indptr_offset + i for i in u_constraints.indptr[1:]]

    rows = dims + u_constraints.shape[0]
    cols = dims + b_mat.shape[1]
    combined_csr = csr_matrix((data, indices, indptr), shape=(rows, cols), dtype=float)

    # from_constraints assumes left-most variables are current-time variables
    return from_constraints(combined_csr, combined_rhs, mode, types=combined_types, names=combined_names, dims=dims)



# Based on:
# https://betterscientificsoftware.github.io/
# python-for-hpc/tutorials/python-pypi-packaging/

from setuptools import setup
import versioneer

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="hylaa",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="Hylaa",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/stanleybak/hylaa",
    author="Stanley Bak",
    author_email="bak2007-DONTSENDMEEMAIL@gmail.com",
    license="GPLv3",
    packages=["hylaa"],
    install_requires=[
        "ffmpeg-python",
        "pytest",
        "numpy",
        "scipy",
        "sympy",
        "matplotlib",
        "termcolor",
        "swiglpk",
        "graphviz",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering"
    ],
)



'''
Stanley Bak
Aggregation Directed Acyclic Graph (DAG) implementation
'''

from termcolor import cprint

from graphviz import Digraph

from hylaa.settings import HylaaSettings
from hylaa.util import Freezable
from hylaa.stateset import StateSet
from hylaa.timerutil import Timers
from hylaa import lputil, aggregate
from hylaa.deaggregation import DeaggregationManager, OpInvIntersect, OpLeftInvariant, OpTransition

class AggDag(Freezable):
    'Aggregation directed acyclic graph (DAG) used to manage the deaggregation process'

    def __init__(self, settings, core):
        self.settings = settings
        self.core = core
        
        self.roots = [] # list of root AggDagNode where the computation begins
        self.cur_node = None # the aggdag_node currently under a continuous post operation

        self.waiting_list = [] # a list of OpTransition (with child_node = None). StateSet is in op.poststate

        self.deagg_man = DeaggregationManager(self)

        self.viz_count = 0
        
        self.freeze_attrs()

    @staticmethod
    def is_op_transition(op):
        'is the passed in op a OpTransition?'

        return isinstance(op, OpTransition)

    def get_cur_state(self):
        '''get the current state being propagated

        This may be None, or a StateSet
        '''

        rv = None

        if self.cur_node is not None:
            rv = self.cur_node.get_cur_state()

        return rv

    def cur_state_left_invariant(self, reached_time_bound=False):
        '''called when the current state left the invariant (or exceeds reach time bound)'''

        state = self.get_cur_state()

        op = OpLeftInvariant(state.cur_step_in_mode, self.cur_node, reached_time_bound)
        self.cur_node.op_list.append(op)

        self.cur_node = None

    def add_init_state(self, state):
        'add an initial state'

        op = OpTransition(0, None, None, None, state)

        self.waiting_list.append(op)

    def print_waiting_list(self):
        'print out the waiting list'

        if self.settings.stdout >= HylaaSettings.STDOUT_VERBOSE:
            col = self.settings.stdout_colors[HylaaSettings.STDOUT_VERBOSE]
            
            cprint("Waiting list has {} states".format(len(self.waiting_list)), col)

            if len(self.waiting_list) < 20:
                for op in self.waiting_list:
                    assert isinstance(op, OpTransition), f"expected OpTransition, but instead had: {op}"
                            
                    state = op.poststate
                    trans = ''

                    if isinstance(op, OpTransition):
                        trans = op.transition
                    
                    cprint(" [Mode: {} (from transition {}) at steps {}]".format(state.mode.name, trans,
                                                                                 state.cur_steps_since_start), col)

    def add_invariant_op(self, step, i_index, is_stronger):
        '''
        an invariant intersection occured, add the invariant intersection op to the current node's op list
        '''

        # OpInvIntersect = namedtuple('OpInvIntersect', ['step', 'node', 'i_index', 'is_stronger'])

        op = OpInvIntersect(step, self.cur_node, i_index, is_stronger)
        self.cur_node.op_list.append(op)

    def make_op_transition(self, t, t_lpi, state, parent_node):
        'make an OpTransition object, can return null if lp solving fails'

        step_in_mode = state.cur_step_in_mode
        steps_since_start = state.cur_steps_since_start
        is_concrete = state.is_concrete

        successor_has_inputs = t.to_mode.b_csr is not None

        op = OpTransition(step_in_mode, parent_node, None, t, None)

        succeeded = self.settings.aggstrat.pretransition(t, t_lpi, op)

        if not succeeded:
            self.core.print_verbose(f"Warning: aggstrat.pretransition returned None (LP solving failed)")
            op = None
        else:
            lputil.add_reset_variables(t_lpi, t.to_mode.mode_id, t.transition_index, \
                reset_csr=t.reset_csr, minkowski_csr=t.reset_minkowski_csr, \
                minkowski_constraints_csr=t.reset_minkowski_constraints_csr, \
                minkowski_constraints_rhs=t.reset_minkowski_constraints_rhs, successor_has_inputs=successor_has_inputs)

            if not t_lpi.is_feasible():
                raise RuntimeError("cur_state became infeasible after reset was applied")

            op_list = [op]
            state = StateSet(t_lpi, t.to_mode, steps_since_start, op_list, is_concrete)
            op.poststate = state

        return op

    def add_transition_successor(self, t, t_lpi, cur_state=None, cur_node=None):
        '''take the passed-in transition from the current state (add to the waiting list)

        t is the AutomatonTransition onject
        t_lpi is an LpInstance consisting of the states in the premode where the current set and guard intersect
        '''

        if cur_state is None:
            cur_state = self.get_cur_state()

        if cur_node is None:
            cur_node = self.cur_node

        op = self.make_op_transition(t, t_lpi, cur_state, cur_node)

        if op:
            cur_node.op_list.append(op)
            self.waiting_list.append(op)

    def _get_node_leaf_ops(self, node):
        'recursively get all the leaf ops originating from the given node'

        rv = []

        child_nodes = []

        for op in node.op_list:
            if isinstance(op, OpTransition):
                if op.child_node is None:
                    rv.append(op)
                elif not op.child_node in child_nodes: # don't insert the same node twice
                    child_nodes.append(op.child_node)

        for child_node in child_nodes:
            rv += self._get_node_leaf_ops(child_node)

        return rv

    def remove_node_decendants_from_waiting_list(self, node):
        'remove all waiting list states originating from the passed-in node'

        to_remove_ops = self._get_node_leaf_ops(node)

        new_waiting_list = []

        for op in self.waiting_list:
            if not op in to_remove_ops:
                new_waiting_list.append(op)

        self.waiting_list = new_waiting_list

    def pop_waiting_list(self):
        'pop a state off the waiting list, possibly doing state-set aggregation'

        op_list = self.settings.aggstrat.pop_waiting_list(self.waiting_list)

        assert op_list, "pop_waiting_list should return non-empty list"
        assert isinstance(op_list[0], OpTransition), f"expected OpTransition, but instead popped: {op_list[0]}"

        # if we're popping error mode, pop them one at a time
        # this is to prevent aggregation, which may use successor mode dynamics information
        if op_list[0].poststate.mode.a_csr is None: 
            op_list = [op_list[0]]

        # if we're popping a mode with no predecessor, an initial mode, pop one at a time
        for op in op_list:
            if op.parent_node is None:
                op_list = [op]
                break

        # remove each element of opr_list from the waiting_list
        waiting_list_presize = len(self.waiting_list)
        self.waiting_list = [op for op in self.waiting_list if not op in op_list]
        assert waiting_list_presize == len(self.waiting_list) + len(op_list), "pop_waiting_list returned " + \
                                                                            "elements not orignally in waiting_list"

        agg_type = None

        if len(op_list) > 1:
            agg_type = self.settings.aggstrat.get_agg_type(op_list)

        # create a new AggDagNode for the current computation
        self.cur_node = self.make_node(op_list, agg_type, 'full')

        return self.cur_node.get_cur_state()

    def make_node(self, ops, agg_type, aggstring):
        '''make an aggdag node

        aggstring is a string that describes the current aggregation from the previous transition
        'full' -> full aggergation
        'init' -> from initial state
        '0100' -> transition was split four times, this is the state from the first, second, first, first partitions
        '''

        node = AggDagNode(ops, agg_type, self, aggstring)

        #if plot:
        #    print(".aggdag make_node() plotting aggdagnode")
        #    self.core.plotman.add_plotted_states([node.stateset])
            
        return node

    def save_viz(self):
        '''save the viz to a sequentially-named file, returns the filename'''

        filename = f"aggdag_{self.viz_count:02d}"
        self.viz_count += 1
                
        self.viz(filename=filename)

        return filename

    def viz(self, lr=True, filename=None):
        'visualize the aggdag using graphviz'

        if filename is not None:
            g = Digraph(name='aggdag', format='png')
        else:
            g = Digraph(name='aggdag')

        if lr:
            g.graph_attr['rankdir'] = 'LR'

        #g.edge_attr.update(arrowhead='dot', arrowsize='2')
        
        already_drawn_nodes = []

        for i, root in enumerate(self.roots):
            preroot = 'preroot{}'.format(i)
            g.node(preroot, style="invis")

            name = "node_{}".format(id(root))
            g.edge(preroot, name)

            root.viz(g, already_drawn_nodes)

        if filename is not None:
            print(f"vizzed to filename: {filename}")
            g.render(filename, cleanup=True)
        else:
            g.view(cleanup=True)

        # sanity check: all nodes in waiting list were drawn
        for op in self.waiting_list:
            if op.parent_node is not None:
                assert op.parent_node in already_drawn_nodes, f"parent node {op.parent_node} not found in drawn nodes"

class AggDagNode(Freezable):
    'A node of the Aggregation DAG'

    def __init__(self, parent_op_list, agg_type, aggdag, aggstring):
        self.aggdag = aggdag
        self.op_list = [] # list of Op* objects
        self.stateset = None # StateSet

        # parent information
        self.parent_ops = parent_op_list
        self.agg_type_from_parents = agg_type

        state_list = [op.poststate for op in parent_op_list]

        # make the stateset
        state = None
        
        if len(state_list) == 1:
            state = state_list[0]
        else:
            self.aggdag.core.print_verbose(f"Aggregating {len(state_list)} states")
            state = self._aggregate_from_state_op_list(state_list, parent_op_list, agg_type)

        state.aggstring = aggstring

        add_root = False

        for op in parent_op_list:
            # child node here may be None, or an existing child node that we override
            op.child_node = self
            
            if op.parent_node is None:
                add_root = True
            
        if add_root:
            self.aggdag.roots.append(self)

        self.stateset = state
        
        self.freeze_attrs()

    # override __hash__ and __eq__ so nodes can be keys in a dict
    def __hash__(self):
        return id(self)

    def __eq__(self, other):
        return self is other

    def __str__(self):
        return f"[AggDagNode in mode {self.stateset.mode.name} w/{len(self.parent_ops)} parent_ops, id={id(self)}]"

    def split(self):
        'for deaggreagtion, split the current (aggregated) node into two, returns a pair of nodes'

        assert len(self.parent_ops) > 1, "attempted to split() unaggregated aggdag node"

        parent_aggstring = self.stateset.aggstring

        if parent_aggstring == 'full':
            parent_aggstring = ''

        rv = []
        
        mid_index = len(self.parent_ops) // 2
        parent_op_lists = [self.parent_ops[:mid_index], self.parent_ops[mid_index:]]

        for agg_suffix, parent_op_list in zip(['0', '1'], parent_op_lists):
            aggstring = parent_aggstring + agg_suffix
            node = self.aggdag.make_node(parent_op_list, self.agg_type_from_parents, aggstring)

            rv.append(node)

        return rv

    def _aggregate_from_state_op_list(self, agg_list, op_list, agg_type):
        '''aggregate states into a single state from a list of states and operatons

        agg_list - a list of states
        op_list - a list of operations
        agg_type the AggType tuple, the aggregation type to use

        returns a single StateSet which is the desired aggregation
        '''

        at = agg_type

        if at.is_chull:
            assert not at.is_box and not at.is_arnoldi_box

            rv = aggregate.aggregate_chull(agg_list, op_list, self.aggdag.core.print_debug)

        elif at.is_box or at.is_arnoldi_box:
            rv = aggregate.aggregate_box_arnoldi(agg_list, op_list, at.is_box, at.is_arnoldi_box, at.add_guard,
                                                 self.aggdag.core.print_debug)
        else:
            raise RuntimeError(f"Unsupported aggregation type: {at}")

        return rv

    def node_left_invariant(self):
        '''
        has the stateset represented by the node become infeasible (left the invariant?)
        '''

        rv = False

        if self.op_list:
            rv = isinstance(self.op_list[-1], OpLeftInvariant)

        return rv

    def replay_op(self, op_list, i):
        '''
        replay a single operation in the current node
        this is used when nodes are split, to leverage parent information
        '''

        Timers.tic('replay_op')

        assert not self.node_left_invariant()

        op = op_list[i]

        cur_state = self.get_cur_state()
        assert cur_state is not None

        self.aggdag.core.print_verbose(f"replaying {i}: {op}")

        if isinstance(op, OpLeftInvariant):
            self.op_list.append(op)
        elif isinstance(op, OpTransition):
            self.replay_op_transition(cur_state, op)
        elif isinstance(op, OpInvIntersect):
            # if there is a later invariant intersection with the same hyperplane and it's stronger, skip this one

            skip = False

            for future_i in range(i+1, len(op_list)):
                future_op = op_list[future_i]

                if not isinstance(future_op, OpInvIntersect):
                    # another op (such as a transition, can't skip current one)
                    break

                if future_op.i_index == op.i_index:
                    if future_op.is_stronger:
                        skip = True

                    break

            #print(".aggdag %%%% debug never skipping invariant intersection")
            # hmm, skipping invariant intersections is only valid if deaggregation is a subset of aggregated set...
            # for our krylov aggregation, this isn't really the case though... so in general you need to invaraint
            # intersect at every step... plus this is really an optimization rather than the main algorithm
            
            if True or not skip:
                self.aggdag.core.print_verbose(
                    f"doing invariant intersection in replay at step {cur_state.cur_step_in_mode}")
                
                is_feasible = self.replay_op_intersect_invariant(cur_state, op)

                if not is_feasible:
                    op = OpLeftInvariant(op.step, self, False)
                    self.op_list.append(op)
            else:
                self.aggdag.core.print_verbose("skipping invariant intersection because stronger one is coming up")

        Timers.toc('replay_op')

    def replay_op_transition(self, state, op):
        '''replay a single operation of type OpTransition
        '''

        print_verbose = self.aggdag.core.print_verbose

        state.step(op.step) # advance the state first

        # check if the transition is still enabled
        
        t = op.transition

        t_lpi = t.get_guard_intersection(state.lpi)

        if t_lpi:
            if t.to_mode.is_error():
                self.aggdag.core.error_reached(state, t, t_lpi)

            new_op = self.aggdag.make_op_transition(t, t_lpi, state, self)

            if not new_op: # make_op_transition can fail due to numerical issues
                print_verbose("Replay Transition {} became infeasible after changing opt direction, skipping")
            else:
                self.op_list.append(new_op)

                if op.child_node is None:
                    self.aggdag.waiting_list.append(new_op)
                    print_verbose("Replay Transition {} when deaggreaged to steps {}".format( \
                                  t, state.cur_steps_since_start))
                else:
                    self.aggdag.deagg_man.update_transition_successors(op, new_op)

                    print_verbose("Replay Transition refined transition {} when deaggregated at steps {}".format( \
                                  t, state.cur_steps_since_start))
        else:
            print_verbose(f"Replay skipped transition {t} when deaggregated to steps {state.cur_steps_since_start}")

    def replay_op_intersect_invariant(self, state, op):
        '''replay a single operation of type OpInvIntersect

        This returns a boolean: is the current state set is still feasible?
        '''

        step, _, invariant_index, is_stronger = op

        state.step(step)

        lc = state.mode.inv_list[invariant_index]

        has_intersection = lputil.check_intersection(state.lpi, lc.negate())
        
        if has_intersection is None:
            rv = False # not feasible
        elif not has_intersection:
            rv = True # no intersection and still feasible
        else: 
            old_row = state.invariant_constraint_rows[invariant_index]
            vec = lc.csr.toarray()[0]
            rhs = lc.rhs

            if old_row is None:
                # new constraint
                row = lputil.add_init_constraint(state.lpi, vec, rhs, state.basis_matrix,
                                                 state.input_effects_list)
                state.invariant_constraint_rows[invariant_index] = row
                is_stronger = False
            else:
                # strengthen existing constraint possibly
                row, is_stronger = lputil.try_replace_init_constraint(state.lpi, old_row, vec, rhs, \
                    state.basis_matrix, state.input_effects_list)
                state.invariant_constraint_rows[invariant_index] = row

            # ad the op to the aggdag
            op = OpInvIntersect(step, self, invariant_index, is_stronger)
            self.op_list.append(op)

            rv = state.lpi.is_feasible()

        return rv

    def get_cur_state(self):
        'get the current state for this node (None if it has left the invariant)'

        if self.op_list and isinstance(self.op_list[-1], OpLeftInvariant):
            rv = None
        else:
            rv = self.stateset

        return rv

    def viz(self, g, already_drawn_nodes):
        '''draw the aggdag node using the graphiz library

        g is the DiGraph object to draw to
        '''

        # sanity check that all parent nodes match
        for parent_op in self.parent_ops:
            assert parent_op.child_node is self

        already_drawn_nodes.append(self)
        name = "node_{}".format(id(self))
        label = self.stateset.mode.name

        if self.op_list and isinstance(self.op_list[-1], OpLeftInvariant):
            steps = self.op_list[-1].step

            if self.op_list[-1].reached_time_bound:
                label += f" ({steps}*)"
            else:
                label += f" ({steps})"
        else:
            steps = self.stateset.cur_step_in_mode
            label += f" (incomplete, {steps} so far)"

        print(f"vizing node {name} ({label}) with {len(self.parent_ops)} parent ops")
        g.node(name, label=label)

        # collapse edges over multiple times into the same outgoing edge
        enabled_transitions = [] # contains tuples (child_name, op_list)

        for op in self.op_list:
            if isinstance(op, OpTransition):
                if op.child_node is None:
                    # unprocessed transition child node
                    child_name = f"out_{id(self)}_{id(op.transition)}"
                else:
                    child_name = "node_{}".format(id(op.child_node))
                    
                found = False

                for pair in enabled_transitions:
                    if pair[0] == child_name:
                        pair[1].append(op)
                        found = True

                if not found:
                    enabled_transitions.append((child_name, [op]))
            elif isinstance(op, OpLeftInvariant):
                # flush remaining enabled_transitions
                for child_name, op_list in enabled_transitions:
                    if len(op_list) == 1:
                        label = str(op_list[0].step)
                    else:
                        label = "[{}, {}]".format(op_list[0].step, op_list[-1].step)
                        

                    if child_name.startswith('out_'): # outgoing edges to unprocessed nodes
                        to_mode_name = op_list[0].transition.to_mode.name
                        g.node(child_name, label=to_mode_name, style="dashed")
                    
                    g.edge(name, child_name, label=label)

                enabled_transitions = []
                # print left-invariant edge
                #invis_name = "out_{}".format(id(op))
                #g.node(invis_name, style="invis")

                #label = str(op.step)
                #g.edge(name, invis_name, label=label, arrowhead='dot', style='dotted')

            ###############################
            # remove enabled transitions that are no longer enabled
            new_enabled_transitions = []

            for pair in enabled_transitions:
                child_name = pair[0]
                op_list = pair[1]

                if op_list[-1].step < op.step - 1:
                    # print it
                    if len(op_list) == 1:
                        label = str(op_list[0].step)
                    else:
                        label = "[{}, {}]".format(op_list[0].step, op_list[-1].step)

                    if child_name.startswith('out_'): # outgoing edges to unprocessed nodes
                        to_mode_name = op_list[0].transition.to_mode.name
                        g.node(child_name, label=to_mode_name, style="dashed")
                        
                    g.edge(name, child_name, label=label)
                else:
                    # keep it
                    new_enabled_transitions.append(pair)

            enabled_transitions = new_enabled_transitions

        # print the children after the current node
        for op in [op for op in self.op_list if isinstance(op, OpTransition) and op.child_node is not None]:
            if not op.child_node in already_drawn_nodes:
                op.child_node.viz(g, already_drawn_nodes)

        # ['step', 'parent_node', 'child_node', 'transition', 'premode_center', 'postmode_state']



'''
Time Elapse Computation. This module is primarily responsive for computing
l * e^{At} where l is some direction of interest, and t is a multiple of some time step
'''

import numpy as np

from hylaa.util import Freezable
from hylaa.timerutil import Timers
from hylaa.time_elapse_expm import TimeElapseExpmMult

class TimeElapser(Freezable):
    'Object which computes the time-elapse function for a single mode at multiples of the time step'

    def __init__(self, mode, step_size):
        self.mode = mode
        self.step_size = step_size
        self.dims = self.mode.a_csr.shape[0]
        self.inputs = 0 if self.mode.b_csr is None else self.mode.b_csr.shape[1]

        self.time_elapse_obj = None

        self.freeze_attrs()

    def get_basis_matrix(self, step_num):
        '''perform the computation for the the basis matrix and input effects matrix at the passed-in step

        returns a tuple (basis_matrix, input_effects matrix)
        '''

        if self.time_elapse_obj is None:
            Timers.tic('init time_elapse_obj')
            self.time_elapse_obj = TimeElapseExpmMult(self)
            Timers.toc('init time_elapse_obj')

        Timers.tic('step')
        self.time_elapse_obj.assign_basis_matrix(step_num)
        Timers.toc('step')

        basis_mat = self.time_elapse_obj.cur_basis_matrix
        input_effects_mat = self.time_elapse_obj.cur_input_effects_matrix

        # post-conditions check
        assert isinstance(basis_mat, np.ndarray), "cur_basis_mat should be an np.array, " + \
            "but it was {}".format(type(basis_mat))

        assert basis_mat.shape == (self.dims, self.dims), \
            "cur_basis mat shape({}) should be {}".format(basis_mat.shape, (self.dims, self.dims))

        if self.inputs == 0 or step_num == 0: # 0-th step input should be null
            assert input_effects_mat is None
        else:
            assert isinstance(input_effects_mat, np.ndarray)

            if self.time_elapse_obj.use_lgg:
                assert input_effects_mat.shape == (self.dims, self.inputs + self.dims)
            else:
                assert input_effects_mat.shape == (self.dims, self.inputs)

        return basis_mat, input_effects_mat

    def use_lgg_approx(self):
        '''
        set this time elapse object to use lgg approximation model
        '''

        self.time_elapse_obj.use_lgg_approx()



'''
Functions related to Kamenev's method for polytope approximation (the method of refined bounds)

Stanley Bak
May 16, 2019
'''

import math

import numpy as np
import scipy as sp
from scipy.spatial import ConvexHull

from hylaa.timerutil import Timers

def _get_orthonormal_rank(vecs, tol=1e-7):
    '''
    given a list of vecs, return a new vector orthonormal to them and the rank of the matrix
    '''

    _, s, v = np.linalg.svd(vecs)

    index = 0

    while index < len(s) and s[index] > tol:
        index += 1

    if index == len(v):
        rv_vec = None # the vectors span the space
    else:
        rv_vec = v[index]

    return rv_vec, index

def _get_rank(vecs, tol=1e-7):
    '''get the rank of the passed in matrix'''

    return _get_orthonormal_rank(vecs, tol=tol)[1]

def _find_two_points(dims, supp_point_func):
    '''find two points in the the convex set defined through supp_point_func (which may be degenerate)

    if len(pts) == 1, the convex set is a degenerate set consisting of a single pt
    '''

    pts = []

    for d in range(dims):
        vec = np.array([-1 if i == d else 0 for i in range(dims)], dtype=float)

        # try min
        p1 = supp_point_func(vec)
        assert len(p1) == dims, f"support fuction returned {len(p1)}-dimensional point, expected {dims}-d"

        pts = [p1]

        # try max
        vec = np.array([1 if i == d else 0 for i in range(dims)], dtype=float)
        p2 = supp_point_func(vec)

        if not np.allclose(p1, p2):
            pts = [p1, p2]
            break

    return pts

def _find_init_simplex(dims, supp_point_func):
    '''
    find an n-dimensional initial simplex
    '''

    Timers.tic('init_simplex')

    # first, construct the initial simplex and determine a basis for the convex set (it may be degenerate)
    init_simplex = _find_two_points(dims, supp_point_func)

    if len(init_simplex) == 2: # S may be a degenerate shape consisting of a single point
        init_vec = init_simplex[1] - init_simplex[0]

        spanning_dirs = [init_vec]
        degenerate_dirs = []
        vecs = [init_vec]

        for _ in range(dims - 1):
            new_dir, rank = _get_orthonormal_rank(vecs)

            # min/max in direction v, checking if it increases the rank of vecs
            pt = supp_point_func(new_dir)
            vecs.append(pt - init_simplex[0])

            if _get_rank(vecs) > rank:
                init_simplex.append(pt)
                spanning_dirs.append(vecs[-1])
                continue

            # rank did not increase with maximize, try minimize
            vecs = vecs[0:-1] # pop vec

            pt = supp_point_func(-1 * new_dir)
            vecs.append(pt - init_simplex[0])

            if _get_rank(vecs) > rank:
                init_simplex.append(pt)
                spanning_dirs.append(vecs[-1])
                continue

            # rank still didn't increase, new_dir is orthogonal to shape S
            vecs = vecs[0:-1] # pop vec

            vecs.append(new_dir) # forces a new orthonormal direction during the next iteration
            degenerate_dirs.append(new_dir)

    Timers.toc('init_simplex')

    return init_simplex

def get_verts(dims, supp_point_func, epsilon=1e-7):
    '''
    get the n-dimensional vertices of the convex set defined through supp_point_func (which may be degenerate)
    '''

    init_simplex = _find_init_simplex(dims, supp_point_func)

    if len(init_simplex) < 3:
        return init_simplex # for 0-d and 1-d sets, the init_simplex corners are the only possible extreme points
    
    rv, _ = _v_h_rep_given_init_simplex(init_simplex, supp_point_func, epsilon=epsilon)

    return rv

def _v_h_rep_given_init_simplex(init_simplex, supp_point_func, epsilon=1e-7):
    '''get all the vertices and hyperplanes of (an epsilon approximation of) the set, defined through supp_point_func

    This function is provided with an initial simplex which spans the space

    this returns verts, equations, where equations is from the Convex Hull's (hull.equations)
    '''

    new_pts = init_simplex
        
    verts = []
    iteration = 0
    max_error = None

    while new_pts:
        iteration += 1
        #print(f"\nIteration {iteration}. Verts: {len(verts)}, new_pts: {len(new_pts)}, max_error: {max_error}")
                
        first_new_index = len(verts)
        verts += new_pts
        new_pts = []
        max_error = 0

        Timers.tic('ConvexHull')
        hull = ConvexHull(verts)
        Timers.toc('ConvexHull')

        for i, simplex in enumerate(hull.simplices):
            is_new = False

            for index in simplex:
                if index >= first_new_index:
                    is_new = True
                    break

            if not is_new:
                continue # skip this simplex

            # get hyperplane for simplex
            normal = hull.equations[i, :-1]
            rhs = -1 * hull.equations[i, -1]

            Timers.tic('supp_point_func')
            supporting_pt = supp_point_func(normal)
            Timers.toc('supp_point_func')
            
            error = np.dot(supporting_pt, normal) - rhs
            max_error = max(max_error, error)

            if error <= -1e-4:
                print(f"Kamenev Plot warning: supporting point was inside facet? error was {error}")

            if error >= epsilon:
                # add the point
                already_added = False

                Timers.tic("check if new point")
                for pt in new_pts:
                    if np.allclose(pt, supporting_pt):
                        already_added = True
                        break
                Timers.toc("check if new point")

                if not already_added:
                    new_pts.append(supporting_pt)

    return np.array(verts, dtype=float), hull.equations



'''
Stanley Bak

Code assocaited with counterexamples.
'''

from collections import defaultdict
from collections import deque

import numpy as np

from sympy import Polygon, Point, Point2D

from scipy.integrate import odeint
from scipy.sparse import csr_matrix

from hylaa.aggstrat import get_ancestors
from hylaa.util import Freezable

class HylaaResult(Freezable): # pylint: disable=too-few-public-methods
    'result object returned by core.run()'

    def __init__(self):
        self.top_level_timer = None # TimerData for total time

        # verification result:
        self.has_aggregated_error = False
        self.has_concrete_error = False

        self.counterexample = [] # if unsafe, a list of CounterExampleSegment objects

        # assigned if setting.plot.store_plot_result is True, an instance of PlotData
        self.plot_data = None

        # the last core.cur_state object... used for unit testing
        self.last_cur_state = None

        # lines plotted for simulations
        # nested lists. first index is plot number, second is segment number,
        # third is simulation number, last is list of 2-d points 
        self.sim_lines = None 

        self.freeze_attrs()

class PlotData(Freezable):
    'used if setting.plot.store_plot_result is True, stores data about the plots'

    def __init__(self, num_plots):
        # first index is plot number, second is mode name, result is a list of obj
        # each obj is a tuple: (verts, state, cur_step_in_mode, str_description)
        self.mode_to_obj_list = [defaultdict(list) for _ in range(num_plots)]

        self.freeze_attrs()

    def get_verts_list(self, mode_name, plot_index=0):
        'get a list of lists for the passed-in mode'

        return self.mode_to_obj_list[plot_index][mode_name][0]

    def add_state(self, state, verts, plot_index):
        'add a plotted state'

        mode_name = state.mode.name

        obj = (verts, state, state.cur_step_in_mode, f"{mode_name} at step {state.cur_step_in_mode}")

        self.mode_to_obj_list[plot_index][mode_name].append(obj)

    def remove_state(self, state, step):
        'remove a state that was previously added'

        found = False

        for mode_to_obj in self.mode_to_obj_list:
            obj_list = mode_to_obj[state.mode.name]

            for index, obj in enumerate(obj_list):
                _, istate, istep, desc = obj

                if istate is state and step == istep:
                    found = True

                    obj_list.pop(index)
                    break

            if found:
                break

        assert found

    def get_plot_data(self, x, y, subplot=0):
        'get the plot data at x, y, or None if not found'

        rv = None

        mode_to_obj = self.mode_to_obj_list[subplot]
        clicked = Point(x, y)

        for mode, obj_list in mode_to_obj.items():
            
            for obj in obj_list:
                verts = obj[0]

                if len(verts) < 4: # need at least 3 points (4 with wrap) to be clicked inside
                    continue

                verts_2dp = [Point2D(x, y) for x, y in verts[1:]]
                                
                poly = Polygon(*verts_2dp)

                if poly.encloses_point(clicked):
                    rv = obj
                    break

            if rv is not None:
                break

        return rv

class CounterExampleSegment(Freezable):
    'a part of a counter-example trace'

    def __init__(self):
        self.mode = None # Mode object
        self.start = []
        self.end = []
        self.steps = 0
        self.outgoing_transition = None # Transition object
        self.reset_minkowski_vars = [] # a list of minkowski variables in the outgoing reset

        self.inputs = deque() # inputs at each step (a deque of m-tuples, where m is the number of inputs)
        
        self.freeze_attrs()

    def __str__(self):
        return "[CE_Segment: {} -> {} in '{}']".format( \
            self.start, self.end, "<None>" if self.mode is None else self.mode.name)

    def __repr__(self):
        return str(self)

def make_counterexample(ha, state, transition_to_error, lpi):
    '''make and return the result counter-example from the lp solution'''

    lp_solution = lpi.minimize() # resolves the LP to get the full unsafe solution
    names = lpi.get_names()

    # first get the number of steps in each mode
    num_steps = []

    if state.aggdag_op_list[0] is not None:
        for node in get_ancestors(state.aggdag_op_list[0].child_node)[1:]:
            assert len(node.parent_ops) == 1
            num_steps.append(node.parent_ops[0].step)

    num_steps.append(state.cur_step_in_mode) # add the last mode

    counterexample = []

    for name, value in zip(names, lp_solution):

        # if first initial variable of mode then assign the segment.mode variable
        if name.startswith('m') and '_i0' in name:
            seg = CounterExampleSegment()
            seg.steps = num_steps[len(counterexample)]
            counterexample.append(seg)

            parts = name.split('_')

            if len(parts) == 2:
                assert len(counterexample) == 1, "only the initial mode should have no predecessor transition"
            else:
                assert len(parts) == 3

                # assign outgoing transition of previous counterexample segment
                transition_index = int(parts[2][1:])
                t = counterexample[-2].mode.transitions[transition_index]
                counterexample[-2].outgoing_transition = t

            mode_id = int(parts[0][1:])

            for mode in ha.modes.values():
                if mode.mode_id == mode_id:
                    seg.mode = mode
                    break

            assert seg.mode is not None, "mode id {} not found in automaton".format(mode_id)

        if name.startswith('m'): # mode variable
            if '_i' in name:
                seg.start.append(value)
            elif '_c' in name:
                seg.end.append(value)
            elif '_I' in name:
                if '_I0' in name:
                    seg.inputs.appendleft([])
                    
                # inputs are in backwards order due to how the LP is constructed, prepend it
                seg.inputs[0].append(value)

        elif name.startswith('reset'):
            seg.reset_minkowski_vars.append(value)

    # add the final transition which is not encoded in the names of the variables
    seg.outgoing_transition = transition_to_error

    return counterexample


def replay_counterexample(ce_segment_list, ha, settings):
    '''replay the counterexample

    returns a list of points and a list of times
    '''

    rv = []
    all_times = []

    epsilon = 1e-7
    step_size = settings.step_size
    
    for i, segment in enumerate(ce_segment_list):
        inv_list = segment.mode.inv_list
        a_mat = segment.mode.a_csr
        b_mat = segment.mode.b_csr

        assert b_mat is None, "todo: implement counter-example generation with inputs"

        der_func = make_der_func(a_mat, b_mat, [])

        start = segment.start
        times = np.linspace(0, segment.steps * step_size, num=segment.steps)
        tol = 1e-9
        states = odeint(der_func, start, times, col_deriv=True, rtol=tol, atol=tol, mxstep=int(1e7))

        rv += [s for s in states]
        t_offset = 0
        
        if all_times:
            t_offset = all_times[-1]
            
        all_times += [t + t_offset for t in times]

        # check if in invariant up to the last step
        for state in states[:-1]:
            for lc in inv_list:
                lhs = lc.csr * state

                assert lhs <= lc.rhs + epsilon, "invariant became false during replay counterexample"

        assert np.allclose(states[-1], segment.end)
                
        # check that last state -> reset -> first state is correct
        t = segment.outgoing_transition
        lhs = t.guard_csr * states[-1]

        for left, right in zip(lhs, t.guard_rhs):
            assert left <= right + epsilon, "guard was not enabled during replay counterexample"

        assert not t.reset_minkowski_csr, "unimplemented: resets with minkowski sums"

        if t.reset_csr is not None:
            poststate = t.reset_csr * states[-1]

            if i + 1 < len(ce_segment_list):
                next_prestate = ce_segment_list[i+1].start

                assert np.allclose(poststate, next_prestate)
            else:
                rv.append(poststate)
                all_times.append(all_times[-1])

    return rv, all_times

def make_der_func(a_matrix, b_matrix, input_vec):
    'make the derivative function with the given paremeters'

    assert isinstance(a_matrix, csr_matrix)

    if b_matrix is not None:
        assert isinstance(b_matrix, csr_matrix)

    input_vec = np.array(input_vec, dtype=float)

    def der_func(state, _):
        'the constructed derivative function'

        no_input = a_matrix * state
        no_input.shape = (state.shape[0],)

        if b_matrix is None:
            rv = no_input
        else:
            input_effects = b_matrix * input_vec
            input_effects.shape = (state.shape[0],)

            rv = np.add(no_input, input_effects)

        return rv

    return der_func



'''
Timer utility functions for Hylaa. Timers are used for performance analysis and
can be refered to statically using Timers.tic(name) and Timers.toc(name)

Stanley Bak
September 2016
'''

import time

from termcolor import cprint

class TimerData():
    'Performance timer which can be started with tic() and paused with toc()'

    def __init__(self, name, parent):
        assert parent is None or isinstance(parent, TimerData)

        self.name = name
        self.total_secs = 0
        self.num_calls = 0
        self.last_start_time = None

        self.parent = parent # parent TimerData, None for top-level timers
        self.children = [] # a list of child TimerData

    def get_child(self, name):
        'get a child timer with the given name'

        rv = None

        for child in self.children:
            if child.name == name:
                rv = child
                break

        return rv

    def get_children_recursive(self, name):
        'get all decendants with the given name (returns a list of TimerData)'

        rv = []

        if name == self.name:
            rv.append(self)

        for child in self.children:
            rv += child.get_children_recursive(name)

        return rv

    def full_name(self):
        'get the full name of the timer (including ancestors)'

        if self.parent is None:
            return self.name

        return "{}.{}".format(self.parent.full_name(), self.name)

    def tic(self):
        'start the timer'

        #print "Tic({})".format(self.name)

        if self.last_start_time is not None:
            raise RuntimeError("Timer started twice: {}".format(self.name))

        self.num_calls += 1
        self.last_start_time = time.perf_counter()

    def toc(self):
        'stop the timer'

        #print "Toc({})".format(self.name)

        if self.last_start_time is None:
            raise RuntimeError("Timer stopped without being started: {}".format(self.name))

        self.total_secs += time.perf_counter() - self.last_start_time
        self.last_start_time = None

class Timers():
    '''
    a static class for doing timer messuarements. Use
    Timers.tic(name) and Timers.tic(name) to start and stop timers, use
    print_stats to print time statistics
    '''

    top_level_timer = None

    stack = [] # stack of currently-running timers, parents at the start, children at the end

    def __init__(self):
        raise RuntimeError('Timers is a static class; should not be instantiated')

    @staticmethod
    def reset():
        'reset all timers'

        Timers.top_level_timer = None
        Timers.stack = []

    @staticmethod
    def tic(name):
        'start a timer'

        #print("Tic({})".format(name))

        if not Timers.stack:
            top = Timers.top_level_timer

            if top is not None and top.name != name:
                # overwrite old top level timer
                #print("Overwriting old top-level timer {} with new top-level timer {}".format(top.name, name))
                top = Timers.top_level_timer = None

            td = top
        else:
            td = Timers.stack[-1].get_child(name)

        # create timer object if it doesn't exist
        if td is None:
            parent = None if not Timers.stack else Timers.stack[-1]
            td = TimerData(name, parent)

            if not Timers.stack:
                Timers.top_level_timer = td
            else:
                Timers.stack[-1].children.append(td)

        td.tic()
        Timers.stack.append(td)

    @staticmethod
    def toc(name):
        'stop a timer'

        #print("Toc({})".format(name))

        assert Timers.stack[-1].name == name, "Out of order toc(). Expected to first stop timer {}".format(
            Timers.stack[-1].full_name())

        Timers.stack[-1].toc()
        Timers.stack.pop()

    @staticmethod
    def print_stats():
        'print statistics about performance timers to stdout'

        Timers.print_stats_recursive(Timers.top_level_timer, 0, None)

    @staticmethod
    def print_stats_recursive(td, level, total_time):
        'recursively print information about a timer'

        low_threshold = 5.0
        high_threshold = 50.0

        if level == 0:
            total_time = td.total_secs

        percent_total = 100 * td.total_secs / total_time

        if percent_total < low_threshold:
            def print_func(text):
                'below threshold print function'

                return cprint(text, 'grey')
        elif percent_total > high_threshold:
            def print_func(text):
                'above threshold print function'

                return cprint(text, None, attrs=['bold'])
        else:
            def print_func(text):
                'within threshold print function'

                return cprint(text, None)

        if td.last_start_time is not None:
            raise RuntimeError("Timer was never stopped: {}".format(td.name))

        if td.parent is None:
            percent = 100
            percent_str = ""
        else:
            percent = 100 * td.total_secs / td.parent.total_secs
            percent_str = " ({:.1f}%)".format(percent)

        print_func("{}{} Time ({} calls): {:.2f} sec{}".format(" " * level * 2, \
            td.name.capitalize(), td.num_calls, td.total_secs, percent_str))

        total_children_secs = 0

        for child in td.children:
            total_children_secs += child.total_secs

            Timers.print_stats_recursive(child, level + 1, total_time)

        if td.children:
            other = td.total_secs - total_children_secs
            other_percent = 100 * other / td.total_secs

            percent_other = other / total_time

            if percent_other < low_threshold:
                def other_print_func(text):
                    'below threshold print function'

                    return cprint(text, 'grey')
            elif percent_other > high_threshold:
                def other_print_func(text):
                    'above threshold print function'

                    return cprint(text, None, attrs=['bold'])
            else:
                def other_print_func(text):
                    'within threshold print function'

                    return cprint(text, None)

            percent_str = " ({:.1f}%)".format(other_percent)

            other_print_func("{}Other: {:.2f} sec{}".format(" " * (level + 1) * 2, \
                other, percent_str))



''''
Stanley Bak
Aggregation Strategy Classes
'''

from collections import namedtuple

from hylaa import lputil
from hylaa.util import Freezable

# container class for description of how to perform a single aggregation
AggType = namedtuple('AggType', ['is_box', 'is_arnoldi_box', 'is_chull', 'add_guard'])

class AggregationStrategy(Freezable):
    '''Aggregation Strategy parent class

    An aggregation strategy controls the evolution of the AggDag. This is done by choosing which transitions create 
    the current node (aggregation), splitting existing nodes (deaggregation) and deciding if/when to stop the
    current node's continuous-post computation (pseudo-transitions).

    The default implementation in this class will perform no aggregation (pop first transition), never split nodes,
    and never stop the current computation. This means it's exact, but analysis can require an exponential amount of
    computation.
    '''

    def __init__(self):
        self.freeze_attrs()

    def init_aggdag_data(self, aggdag):
        'initialize data within the aggdag object by directly assigning to aggdag instance'

        pass

    def init_aggdag_node_data(self, aggdag_node):
        'initialize data within every aggdag node by directly assigning to the aggdag_node instance'

        pass

    def pop_waiting_list(self, waiting_list):
        '''determine which waiting list elements should be aggregated for the next continuous post computation

        waiting_list is a list of OpTransition

        this function returns a list of OpTransition. 
        If the list is a single element, no aggregation is performed.
        '''

        return [waiting_list[0]]

    def get_deagg_node(self, aggdag):
        '''Called before popping a state off the waiting list. Get the aggdag node to deaggregate (if any).
        '''

        return None

    def pretransition(self, t, t_lpi, op_transition):
        '''event function, called when taking a transition before the reset is applied

        returns True if succeeded, False on LP errors
        '''

        # premode_center = lputil.get_box_center(t_lpi)
        return True

    def get_agg_type(self, op_list):
        '''
        Gets the type of aggregation to be performed for the passed in objects. 

        This returns an instance of AggType
        '''

        raise NotImplementedError("Unaggregated strategy does not implement get_agg_type")

class Unaggregated(AggregationStrategy):
    'another name for the base implementation of AggregationStrategy'

class Aggregated(AggregationStrategy):
    'a fully aggregated strategy'

    AGG_BOX, AGG_ARNOLDI_BOX, AGG_CONVEX_HULL = range(3)
    POP_LOWEST_MINTIME, POP_LOWEST_AVGTIME, POP_LARGEST_MAXTIME = range(3)
    DEAGG_LEAVES_FIRST, DEAGG_ROOT_FIRST, DEAGG_MOST_STATES = range(3)

    def __init__(self, agg_type=AGG_ARNOLDI_BOX, deaggregate=False):
        self.agg_type = agg_type
        self.pop_type = Aggregated.POP_LOWEST_AVGTIME

        self.add_guard = True # add the guard direction when performing aggregation
        self.require_same_path = True
        self.deaggregate = deaggregate
        self.deagg_preference = Aggregated.DEAGG_MOST_STATES

        self.sim_avoid_modes = [] # list of mode names to try to avoid during simulation

        AggregationStrategy.__init__(self)

    def pop_waiting_list(self, waiting_list):
        '''
        Get the states to remove from the waiting list based on a score-based method
        '''

        # use score method to decide which mode to pop
        to_remove_op = waiting_list[0]
        to_remove_state = to_remove_op.poststate
        to_remove_score = self.pop_score(to_remove_state)

        for op in waiting_list:
            state = op.poststate
            score = self.pop_score(state)

            if score > to_remove_score or score == to_remove_score and state.mode.name < to_remove_state.mode.name:
                to_remove_score = score
                to_remove_state = state
                to_remove_op = op

        # remove all states for aggregation
        op_list = []

        for op in waiting_list:
            state = op.poststate
            should_add = False

            if state is to_remove_state:
                should_add = True
            elif state.mode is to_remove_state.mode:
                if not self.require_same_path:
                    should_add = True
                else: # require same path; paths are the same if the parent node and transitions match
                    if to_remove_op is None and op is None:
                        should_add = True
                    elif to_remove_op and op:
                        if to_remove_op.transition is op.transition and \
                                                    to_remove_op.parent_node is op.parent_node:
                            should_add = True

            if should_add:
                op_list.append(op)

        return op_list
    
    def pop_score(self, state):
        '''
        returns a score used to decide which state to pop off the waiting list. The state with the highest score
        will be removed first.
        '''

        if self.pop_type == Aggregated.POP_LOWEST_MINTIME:
            score = -(state.cur_steps_since_start[0])
        elif self.pop_type == Aggregated.POP_LOWEST_AVGTIME:
            score = -(state.cur_steps_since_start[0] + state.cur_steps_since_start[1])
        elif self.pop_type == Aggregated.POP_LARGEST_MAXTIME:
            score = state.cur_steps_since_start[1]
        else:
            raise RuntimeError("Unknown waiting list pop strategy: {}".format(self.pop_type))

        return score

    def get_simulation_pop_mode(self, sim_waiting_list):
        '''
        returns the mode to be popped off the simulation waiting list
        '''

        # find minimum time mode
        min_time_mode = None
        min_time_steps = float('inf')

        for mode, _, steps in sim_waiting_list:
            if mode.name in self.sim_avoid_modes:
                steps = float('inf')

            if steps < min_time_steps:
                min_time_mode = mode
                min_time_steps = steps

        if min_time_mode is None: # all are avoid modes
            min_time_mode = sim_waiting_list[0][0]
                
        return min_time_mode

    def pretransition(self, t, t_lpi, op_transition):
        'event function, called when taking a transition before the reset is applied'

        rv = True

        if self.agg_type == Aggregated.AGG_ARNOLDI_BOX:
            op_transition.premode_center = lputil.get_box_center(t_lpi)

            if op_transition.premode_center is None:
                rv = False

        return rv

    def get_agg_type(self, op_list):
        '''
        Gets the type of aggregation to be performed for the passed in objects. 

        This returns an instance of AggType
        '''

        return self._get_agg_type()

    def _get_agg_type(self):
        '''
        Gets the type of aggregation to be performed.
        '''
    
        is_box = self.agg_type == Aggregated.AGG_BOX
        is_arnoldi_box = self.agg_type == Aggregated.AGG_ARNOLDI_BOX
        is_chull = self.agg_type == Aggregated.AGG_CONVEX_HULL

        return AggType(is_box, is_arnoldi_box, is_chull, self.add_guard)

    def get_deagg_node(self, aggdag):
        '''Called before popping a state off the waiting list. Get the aggdag node to deaggregate (if any).
        '''

        rv = None

        if self.deaggregate:

            # if nodes have multiple outgoing transitions they will become deaggregated
            node_to_transition = {}

            for op in aggdag.waiting_list:
                state = op.poststate

                if state.is_concrete: # only try to split non-concrete states
                    continue

                # highest deaggregation priority: non-concrete states that reach an error mode
                if state.mode.is_error():
                    rv = op.parent_node
                    #break

                # other deaggregation condition: different outgoing transitions from the same node
                if op.parent_node in node_to_transition:
                    transition = node_to_transition[op.parent_node]

                    if op.transition is not transition:
                        rv = op.parent_node
                else:
                    node_to_transition[op.parent_node] = op.transition

                    assert op.parent_node.node_left_invariant()

                    # if it's a single state that looks like it reaches the time bound
                    #if op.parent_node.op_list[-1].reached_time_bound:
                    #    # both a transition and reached time bound... split it
                    #    rv = op.parent_node
                    #    print(f"rv is state with a transition that also reached the time bound")

        # split earlier or latest ancestor, depending on settings
        if rv:
            ancestors = get_ancestors(rv)

            if self.deagg_preference == Aggregated.DEAGG_LEAVES_FIRST:
                ancestors.reverse()
            elif self.deagg_preference == Aggregated.DEAGG_MOST_STATES:
                # sort ancestors by the number of parent states
                count_parent_ops = lambda node: len(node.parent_ops)
                ancestors = reversed(sorted(ancestors, key=count_parent_ops))

            rv = None

            for node in ancestors:
                if len(node.parent_ops) > 1:
                    rv = node
                    break

            assert rv is not None, "didn't find aggregated ancestor?"

        return rv

def get_ancestors(node):
    '''
    get an ordered list of all the ancestors, starting from the root to the leaf
    '''

    rv = []

    if node.parent_ops[0].parent_node is not None:
        rv += get_ancestors(node.parent_ops[0].parent_node)

    rv.append(node)

    return rv
